'''
Module contains a Class that is a scikit learn Transformer.
'''
from sklearn.base import TransformerMixin
from sklearn.base import BaseEstimator

from bella import contexts


class Context(BaseEstimator, TransformerMixin):

    def __init__(self, context='left', inc_target=False):
        self.context = context
        self.inc_target = inc_target

    def fit(self, target_dicts, y=None):
        '''Kept for consistnecy with the TransformerMixin'''

        return self

    def fit_transform(self, target_dicts, y=None):
        '''see self.transform'''

        return self.transform(target_dicts)


    def transform(self, target_dicts):
        '''
        Given a list of target dictionaries containing the spans of the targets
        and the texts that are about the targets it returns the relevant left,
        right and target contexts with respect to the target word(s). Returns a
        list of contexts.

        :param target_dicts: list of dictionaries containing at least `spans` and \
        `text` keys.
        :type target_dicts: list
        :returns: a list of left, right and target contexts with respect to the \
        target word and the values in the self.context if self.context = 'lt' will \
        only return the left and target contexts and not right.
        :rtype: list
        '''

        all_context_data = []
        for target_dict in target_dicts:
            context_data = []
            context_data.extend(contexts.context(target_dict, self.context,
                                                 self.inc_target))
            all_context_data.append(context_data)
        return all_context_data



'''
A list of functions that convert TargetCollections into a specified file format.

1. semeval_14: Converts the TargetCollection into XML file format that was used \
in `SemEval 2014 task 4 <http://alt.qcri.org/semeval2014/task4/>`_
'''
import xml.etree.ElementTree as ET

def semeval_14(file_path, target_collection):
    '''
    :param file_path: Location of the file to save the XML data to
    :param target_collection: The target collection data to convert into the \
    XML file.
    :type file_path: String
    :type target_collection: TargetCollection
    :returns: Nothing. Saves the data to the file path given.
    :rtype: None
    '''
    sentiment_mapper = {1 : 'positive', 0 : 'neutral', -1 : 'negative'}
    sentence_targets = target_collection.group_by_sentence()
    tree = ET.Element('sentences')
    for sentence_id, targets in sentence_targets.items():
        # Removes part of the sentence id we add from our parsing
        if sentence_id.startswith('samsung_galaxy_s5'):
            sentence_id = sentence_id[len('samsung_galaxy_s5'):]
        sentence_element = ET.SubElement(tree, 'sentence',
                                         attrib={'id' : sentence_id})
        text_element = ET.SubElement(sentence_element, 'text')
        text_element.text = targets[0]['text']
        aspects_element = ET.SubElement(sentence_element, 'aspectTerms')
        for target in targets:
            spans = target['spans']
            if len(spans) > 1:
                raise ValueError('There should only be one set of spans not'\
                                 '{}. List of spans: {}'\
                                 .format(len(spans), spans))
            span_from, span_to = spans[0]
            attributes = {'term' : target['target'],
                          'polarity' : sentiment_mapper[target['sentiment']],
                          'from' : str(span_from), 'to' : str(span_to)}
            aspect_element = ET.SubElement(aspects_element, 'aspectTerm',
                                           attrib=attributes)

    with open(file_path, 'w') as xml_file:
        tree = ET.ElementTree(tree)
        tree.write(xml_file, encoding='unicode', xml_declaration=True)



from bella import contexts, data_types, dependency_parsers, dependency_tokens
from bella import evaluation, helper, lexicons, neural_pooling
from bella import parsers, stanford_tools
from bella import syntactic_contexts, tokenisers, word_vectors, write_data
from bella import scikit_features, models



import json

from nltk.tree import Tree
from stanfordcorenlp import StanfordCoreNLP


class StanfordNlp(object):
    '''
    Singleton Class instance
    '''

    instance = None

    def __new__(cls):
        if StanfordNlp.instance is None:
            StanfordNlp.instance = StanfordCoreNLP('http://localhost', 9000)
        return StanfordNlp.instance

    def __getattr__(self, name):
        return getattr(self.instance, name)

    def __setattr__(self, name, value):
        return setattr(self.instance, name, value)


def tokenise(text):

    stanford_nlp = StanfordNlp()
    output_dict = stanford_nlp.annotate(text, {'annotators': 'ssplit,tokenize',
                                               'tokenize.language': 'English',
                                               'outputFormat': 'json'})
    output_dict = json.loads(output_dict, strict=False)
    tokens = [token['originalText'] for s in output_dict['sentences']
              for token in s['tokens']]
    return tokens


def constituency_parse(text):
    '''
    :param text: The text you want to parse
    :type text: String
    :returns: A list of parse trees where each tree is represented by \
    nltk.tree.Tree. Each parse tree is associated to a sentence in the text. \
    If one sentence in the text then the list will be of length 1.
    :rtype: list
    '''

    if text.strip() == '':
        raise ValueError('There has to be some text to parse. Text given {}'
                         .format(text))
    stanford_nlp = StanfordNlp()
    output_dict = stanford_nlp.annotate(text, {'annotators': 'pos,parse',
                                               'tokenize.language': 'English',
                                               'outputFormat': 'json'})
    output_dict = json.loads(output_dict)
    parse_trees = [Tree.fromstring(sent['parse'])
                   for sent in output_dict['sentences']]
    return parse_trees


def dependency_parse(text, dep_type='basicDependencies'):
    '''
    :param text: The text you want to parse
    :param dep_type: The dependency type to use either: 'basicDependencies', \
    'enhancedDependencies', or 'enhancedPlusPlusDependencies'. See for more \
    details \
    https://nlp.stanford.edu/~sebschu/pubs/schuster-manning-lrec2016.pdf
    :type text: String
    :type dep_type: String. Default basicDependencies
    '''

    if text.strip() == '':
        raise ValueError('There has to be some text to parse. Text given {}'
                         .format(text))

    stanford_nlp = StanfordNlp()
    output_dict = stanford_nlp.annotate(text, {'annotators': 'pos,depparse',
                                               'tokenize.language': 'English',
                                               'outputFormat': 'json'})
    sentences = json.loads(output_dict)['sentences']
    tokens_dicts = []
    dep_dicts = []
    for sentence in sentences:
        tokens_dict = {token_data['index']: token_data
                       for token_data in sentence['tokens']}
        dep_dict = {}
        for dep_data in sentence[dep_type]:
            dep_rel = dep_data['dep']
            dep_token_index = dep_data['governor']
            current_token_index = dep_data['dependent']
            dep_dict[current_token_index] = (dep_rel, dep_token_index)
        tokens_dicts.append(tokens_dict)
        dep_dicts.append(dep_dict)
    return dep_dicts, tokens_dicts



'''
Contains the following class:

1. DependencyToken
'''


class DependencyToken():
    '''
    Objects that contain a token and its dependency relations via it's attributes.

    Attributes:

    1. self.token -- The token stored e.g. `hello`
    2. self.relations -- The dict that contains words associated to the token \
    at different dependency relation levels e.g. relation of 1 direct dependency \
    of the token, relation 2 dependency relations of the relation 1 words.
    3. self.connected_words -- A list of tuples (word, relation) where all words \
    are within the same dependency tree and are listed in the order they appear \
    in the text that was parsed. relation can take two values `CURRENT` current \
    token and `CONNECTED` any other token.

    Methods:

    1. get_n_relations -- The lower and upper bound of the range `n` for the \
    words retrived at different relation levels associated to the `n` range.
    '''

    def __init__(self, token, relations, connected_words):
        '''
        :param token: The string that the dependency relations are associated to.
        :param relations: dict of dependency relations
        :param connected_words: list of current token and all of its syntactically\
         connected words where the list is made up of tuples where the first value \
         is the String of the connected word and the second describes its realtion \
         either `CONNECTED` for connected words or `CURRENT` stating it is the \
         current token. All words are ordered by their occurence in the text. \
         The connected words are all words in the same dependecy tree.
        :type token: String
        :type relations: dict
        :type connected_words: list
        :returns: Constructor
        :rtype: None
        '''

        if not isinstance(token, str):
            raise TypeError('token parameter has to be of type str not {}'\
                            .format(type(token)))
        if not isinstance(relations, dict):
            raise TypeError('relations parameter has to be of type dict not {}'\
                            .format(type(relations)))
        if not isinstance(connected_words, list):
            raise TypeError('connected words parameter has to be of type list not'\
                            ' {}'.format(connected_words))
        if len(connected_words) < 1:
            raise ValueError('connected words has to always contain one word which'\
                             ' is the current word')

        current_count = 0
        valid_relations = set(['CONNECTED', 'CURRENT'])
        for connected_word in connected_words:
            if not isinstance(connected_word, tuple):
                raise TypeError('connected_words parameter has to be a list of '\
                                'tuples not {}'.format(type(connected_word)))
            word, relation = connected_word
            if relation not in valid_relations:
                raise ValueError('connected relations can only be {} and not {}'\
                                 .format(valid_relations, relation))
            if not isinstance(word, str):
                raise TypeError('The word in the connected_word has to be of '\
                                'type String and not {}'.format(type(word)))
            if relation == 'CURRENT':
                current_count += 1
                if word != token:
                    raise ValueError('The token that is the CURRENT token {} in '\
                                     'connected_words should also be equal to '\
                                     'the token value {}'.format(word, token))
        if current_count != 1:
            raise ValueError('There has to be ONLY one CURRENT relation in '\
                             'the connected words {}'.format(connected_words))

        all_depths = []
        for depth_index, related_list in relations.items():
            if not isinstance(depth_index, int):
                raise TypeError('The keys in the relations dict have to be of '\
                                'type int not {}'.format(type(depth_index)))
            if not isinstance(related_list, list):
                raise TypeError('The related words in the relations dict have to '\
                                'be of type list and not {}'.format(type(related_list)))
            all_depths.append(depth_index)
        if relations != {}:
            actual_depth_range = sorted(all_depths)
            start_depth = actual_depth_range[0]
            if start_depth != 1:
                raise ValueError('The depth of the relations should always start at '\
                                 '1 and not less or greater than your start depth {}'\
                                 .format(start_depth))
            max_depth = max(all_depths)
            valid_depth_range = list(range(1, max_depth + 1))
            if valid_depth_range != actual_depth_range:
                raise ValueError('The depths (keys) in the relations dict has to be '\
                                 'incremental e.g. 1,2,3,4 and not 1,3,4 the depths '\
                                 'in your relations {} valid depths {}'\
                                 .format(actual_depth_range, valid_depth_range))
        self.token = token
        self.relations = relations
        self.connected_words = connected_words

    def get_n_relations(self, relation_range=(1, 1)):
        '''
        Given a tuple to denote the range of the depth of dependency relations to
        return. Returns a list of words.

        If the relation range is greater than the depth range then it will return
        all words up to the last depth and ignore the depths that do not exisit.

        Negative ranges can be used in a similar manner to how list indexs work.

        :param relation_range: Tuple to denote the range of the depth of \
        dependency relations to return.
        :type relation_range: tuple
        :returns: A list of words that are associated to the token at the depth \
        range given.
        :rtype: list

        :Example:
        >>> rel_token = DependencyToken('example', {1 : ['hello'], 2 : ['anything'],
                                        3 : ['lastly']})
        >>> relations = rel_token.get_n_relations((2,3))
        >>> relations == ['anything', 'lastly']
        >>> relations = rel_token.get_n_relations((-2,-1))
        >>> relations == ['anything', 'lastly']
        '''

        def negative_range(range_1, range_2):
            '''
            Given two integers where one or both are negative returns the relations
            at those range depths.

            :param range_1: start of the range
            :param range_2: end of the range
            :type range_1: int
            :type range_2: int
            :returns: Given two integers where one or both are negative returns \
            the relations at those range depths.
            :rtype: list
            '''

            def flat_list(word_lists):
                '''
                :param word_lists: A list which contains lists of Strings
                :type word_lists: list
                :returns: Converts a nested list of lists into a flat list.
                :rtype: list
                '''

                return [word for words in word_lists for word in words]

            i = 1
            all_words = []
            while True:
                related_words = self.relations.get(i, [])
                if related_words == []:
                    break
                i += 1
                all_words.append(related_words)
            if range_2 == range_1:
                if range_2 == -1:
                    return flat_list(all_words[range_1 :])
                return flat_list(all_words[range_1 : range_2 + 1])
            if range_1 > 0:
                range_1 -= 1
            if range_2 > 0:
                range_2 -= 1
            if range_2 == -1:
                return flat_list(all_words[range_1 :])
            elif range_2 < -1:
                return flat_list(all_words[range_1 : range_2 + 1])
            return flat_list(all_words[range_1 : range_2])

        if not isinstance(relation_range, tuple):
            raise TypeError('The relation_range parameter has to be of type tuple'\
                            ' and not {}'.format(type(relation_range)))
        if len(relation_range) != 2:
            raise ValueError('The relation_range tuple has to be of length 2 '\
                             'and not {}'.format(len(relation_range)))
        range_1 = relation_range[0]
        range_2 = relation_range[1]
        if not (isinstance(range_1, int)
                and isinstance(range_2, int)):
            raise ValueError('The relation_range tuple can only contain value of '\
                             'type int values {}'.format(relation_range))
        if range_1 == 0 or range_2 == 0:
            raise ValueError('relation_range values cannot be zero')
        all_related_words = []
        # Check if it is negative indexing
        if range_1 < 0:
            if range_2 < range_1:
                raise ValueError('If the first value in the range is negative it'\
                                 ' has to be less than the second value in the '\
                                 'range 1: {} 2: {}'.format(range_1, range_2))
            all_related_words = negative_range(range_1, range_2)
        elif range_2 < 0:
            all_related_words = negative_range(range_1, range_2)
        else:

            if range_1 < 1 or range_2 < 1:
                raise ValueError('The values in the relation range have to '\
                                 'be greater than one')
            if range_1 > range_2:
                raise ValueError('The second value in relation_range has to '\
                                 'be greater than or equall to the first value.'\
                                 ' Values {}'.format(relation_range))
            for i in range(range_1, range_2 + 1):
                related_words = self.relations.get(i, [])
                if related_words == []:
                    break
                all_related_words.extend(related_words)
        return all_related_words

    def connected_target_span(self, renormalise=None):
        '''
        :param renormalise: A tuple containing the normalised target and the \
        real target word e.g. (Samsung_S5, Samsung S5). If you do not require \
        this leave as default parameter None.
        :type renormalise: tuple. Default None
        :returns: It returns connected words as a String and the span of the current \
        token within that String as a tuple of length 2.
        :rtype: tuple
        '''

        connected_words = []
        word_length = {}
        target_index = -1
        for index, word_relation in enumerate(self.connected_words):
            word, relation = word_relation
            if renormalise is not None:
                normalised_target, real_word = renormalise
                if word == normalised_target:
                    word = real_word
            connected_words.append(word)
            if relation == 'CURRENT':
                target_index = index
            word_length[index] = len(word) + word_length.get(index - 1, 0)
        if target_index == -1:
            raise ValueError('target index can never be -1 error in the '\
                             'connected_words {}'.format(connected_words))
        connected_text = ' '.join(connected_words).strip()
        num_previous_chars = word_length.get(target_index - 1, 0)
        if target_index != 0:
            num_previous_chars += target_index
        if renormalise is not None:
            self.token = renormalise[1]
        target_word = self.token
        target_span = (num_previous_chars, num_previous_chars + len(target_word))
        if connected_text[target_span[0]: target_span[1]] != target_word:
            raise ValueError('Cannot get the target word `{}` within the connected '\
                             'text `{}` this could be due to the connected words '\
                             '`{}` spans: `{}`'.format(target_word, connected_text,
                                                       connected_words, target_span))
        return connected_text, target_span



'''
Module contains a Class that is a scikit learn Transformer.

Classes:

1. SyntacticContext - Converts a list of dictionaries containg text, targets, \
and target spans into text contexts defined the targets dependency tree. \
Returns a list of a list of dictionaries containing text and span.
2. DependencyChildContext - Simialr to SyntacticContext but returns a list of \
a list of Strings instead of dicts. Each String represents a targets children \
dependency relations.
3. Context - Given the output from SyntacticContext returns the left, right,
target or full text contexts. Where left and right is with respect to the
target word the text is about.
'''

from sklearn.base import TransformerMixin
from sklearn.base import BaseEstimator

from bella import syntactic_contexts
from bella.dependency_parsers import tweebo


class SyntacticContext(BaseEstimator, TransformerMixin):
    '''
    Converts a list of dictionaries containg text, targets, and target spans
    into text contexts defined the targets dependency tree. Returns a list of a
    list of dictionaries containing text and span.

    Attributes:

    1. parser - dependency parser to use.
    2. lower - whether or not the parser should process the text in lower or \
    case or not

    Methods:

    1. fit - Does nothing as nothing is done at fit time.
    2. fit_transform - Performs the transform method.
    3. transform - Converts the list of dicts into a list of a list of \
    dicts where each each dict contains the target text and span where the text \
    is the targets full dependency tree word context.
    '''

    def __init__(self, parser=tweebo, lower=False):
        '''
        For more information on what the function does see the following
        functions documentation: dependency_context

        :param parser: Dependency parser to use.
        :param lower: Whether to lower case the words before going through \
        the dependency parser.
        :type parser: function. Default tweebo
        :type lower: bool. Default False
        '''

        self.parser = parser
        self.lower = lower

    def fit(self, target_dicts, y=None):
        '''Kept for consistnecy with the TransformerMixin'''

        return self

    def fit_transform(self, target_dicts, y=None):
        '''see self.transform'''

        return self.transform(target_dicts)


    def transform(self, target_dicts):
        '''
        Given a list of target dictionaries it returns the syntactic context of
        each target therefore returns a list of a list of dicts where each
        dict represents a targets syntactic context within the associated text.
        The syntactic context depends on the self.parser function.

        :param target_dicts: list of dictionaries
        :type target_dicts: list
        :returns: A list of a list of dicts
        :rtype: list
        '''

        return syntactic_contexts.dependency_context(target_dicts,
                                                     self.parser, self.lower)

class DependencyChildContext(BaseEstimator, TransformerMixin):
    '''
    Simialr to SyntacticContext but returns a list of a list of Strings
    instead of dicts. Each String represents a targets children dependency
    relations.

    Attributes:

    1. parser - dependency parser to use.
    2. lower - whether or not the parser should process the text in lower or \
    case or not
    3. rel_depth - The depth of the child relations (1, 1) = 1st relations

    Methods:

    1. fit - Does nothing as nothing is done at fit time.
    2. fit_transform - Performs the transform method.
    3. transform - Converts the list of dicts into a list of a list of \
    Strings which are the target words child relations within the targets \
    dependency tree.
    '''

    def __init__(self, parser=tweebo, lower=False, rel_depth=(1, 1)):
        '''
        For more information on what the function does see the following
        functions documentation: dependency_relation_context

        :param parser: Dependency parser to use.
        :param lower: Whether to lower case the words before going through \
        the dependency parser.
        :param rel_depth: Depth of the dependency relations to use as context \
        words
        :type parser: function. Default tweebo
        :type lower: bool. Default False
        :type rel_depth: tuple. Default (1, 1)
        '''
        self.parser = parser
        self.lower = lower
        self.rel_depth = rel_depth

    def fit(self, target_dicts, y=None):
        '''Kept for consistnecy with the TransformerMixin'''

        return self

    def fit_transform(self, target_dicts, y=None):
        '''see self.transform'''

        return self.transform(target_dicts)


    def transform(self, target_dicts):
        '''
        Given a list of target dictionaries it returns a list of a list of Strings
        where each String is the concatenation of the targets rel_depth child
        related words where rel_depth determines the number of child dependency
        related words to include.

        :param target_dicts: list of dictionaries
        :type target_dicts: list
        :returns: A list of a list of Strings
        :rtype: list
        '''

        return syntactic_contexts.dependency_relation_context(target_dicts,
                                                              self.parser,
                                                              self.lower,
                                                              self.rel_depth)

class Context(BaseEstimator, TransformerMixin):
    '''
    Given the output from SyntacticContext returns the left, right, target or
    full text contexts. Where left and right is with respect to the
    target word the text is about.

    Attributes:

    1. context - Defines the text context that is returned
    2. inc_target - Whether to include the target word in the context

    Methods:

    1. fit - Does nothing as nothing is done at fit time.
    2. fit_transform - Performs the transform method.
    3. transform - Converts a list of a list of dictionaries into a list of a \
    list of Strings where each String represents a the targets context.
    '''
    def __init__(self, context='left', inc_target=False):
        '''
        :param context: left, right, target or full context will be returned \
        with respect to the target word in the target sentence.
        :param inc_target: Whether to include the target word in the text \
        context.
        :type context: String. Can only be `left`, `right`, `target`, or `full`\
        . Default `left`
        :type inc_target: bool. Default False
        '''
        self.context = context
        self.inc_target = inc_target

    def fit(self, target_contexts, y=None):
        '''Kept for consistnecy with the TransformerMixin'''

        return self

    def fit_transform(self, target_contexts, y=None):
        '''see self.transform'''

        return self.transform(target_contexts)


    def transform(self, target_contexts):
        '''
        Given a list of of a list of dicts where each dict contains

        :param target_dicts: list of a list of dictionaries
        :type target_dicts: list
        :returns: A list of a list of Strings where each String represents a \
        different context as each target can have many targets within a text \
        thus multiple contexts. The context it returns depends on self.context
        :rtype: list
        '''

        return syntactic_contexts.context(target_contexts, self.context,
                                          self.inc_target)



'''
Module contains all of the classes that represent Machine Learning models
that are within `Tang et al. 2016 paper \
<https://aclanthology.info/papers/C16-1311/c16-1311>`_:

1. :py:class:`bella.models.tdlstm.LSTM` -- LSTM model.
2. :py:class:`bella.models.tdlstm.TDLSTM` -- TDLSTM model.
3. :py:class:`bella.models.tdlstm.TCLSTM` -- TCLSTM model.
'''

from typing import Dict, Callable, Any, List, Union, Tuple

import numpy as np
import keras
from keras import models, optimizers, layers
from keras.utils import to_categorical

import bella
from bella.models.base import KerasModel
from bella.contexts import context
from bella.neural_pooling import matrix_median


class LSTM(KerasModel):
    '''
    Attributes:

    1. pad_size -- The max number of tokens to use per sequence. If -1
       use the text sequence in the training data that has the most tokens as
       the pad size.
    2. embedding_layer_kwargs -- Keyword arguments to pass to the embedding
       layer which is a :py:class:`keras.layers.Embedding` object. Can be
       None if no parameters to pass.
    3. lstm_layer_kwargs -- Keyword arguments to pass to the lstm layer(s)
       which is a :py:class:`keras.layers.LSTM` object. Can be
       None if no parameters to pass.
    4. dense_layer_kwargs -- Keyword arguments to pass to the dense (final
       layer) which is a :py:class:`keras.layers.Dense` object. Can be
       None if no parameters to pass.

    Methods:

    1. model_parameters -- Returns a dictionary containing the attributes of
       the class instance, the parameters to give to the class constructior to
       re-create this instance, and the class itself.
    2. create_training_text -- Converts the training and validation data into a
       format that the keras model can take as input.
    3. create_training_y -- Converts the training and validation target values
       from a vector of class lables into a matrix of binary values. of shape
       [n_samples, n_classes].
    4. keras_model -- The model that represents this class. This is a single
       forward LSTM.
    '''

    @classmethod
    def name(cls) -> str:
        return 'LSTM'

    def __repr__(self) -> str:
        '''
        Name of the machine learning model.
        '''
        return self.name()

    def __init__(self, tokeniser: Callable[[str], List[str]],
                 embeddings: 'bella.word_vectors.WordVectors',
                 reproducible: Union[int, None] = None, pad_size: int = -1,
                 lower: bool = True, patience: int = 10,
                 batch_size: int = 32, epochs: int = 300,
                 embedding_layer_kwargs: Dict[str, Any] = None,
                 lstm_layer_kwargs: Dict[str, Any] = None,
                 dense_layer_kwargs: Dict[str, Any] = None,
                 optimiser: 'keras.optimizers.Optimizer' = optimizers.SGD,
                 optimiser_params: Union[Dict[str, Any], None] = None
                 ) -> None:
        '''
        :param tokeniser: Tokeniser to be used e.g. :py:meth:`str.split`.
        :param embeddings: Embedding (Word vectors) to be used e.g.
                           :py:class:`bella.word_vectors.SSWE`
        :param reproducible: Whether to be reproducible. If None then it is
                             quicker to run. Else provide a `int` that
                             will represent the random seed value.
        :param pad_size: The max number of tokens to use per sequence. If -1
                         use the text sequence in the training data that has
                         the most tokens as the pad size.
        :param lower: Whether to lower case the words being processed.
        :param patience: Number of epochs with no improvement before training
                         is stopped.
        :param batch_size: Number of samples per gradient update.
        :param epochs: Number of times to train over the entire training set
                       before stopping. If patience is set, then it may
                       stop before reaching the number of epochs specified
                       here.
        :param embedding_layer_kwargs: Keyword arguments to pass to the
                                       embedding layer which is a
                                       :py:class:`keras.layers.Embedding`
                                       object. If no parameters to pass leave
                                       as None.
        :param lstm_layer_kwargs: Keyword arguments to pass to the lstm
                                  layer(s) which is a
                                  :py:class:`keras.layers.LSTM` object. If no
                                  parameters to pass leave as None.
        :param dense_layer_kwargs: Keyword arguments to pass to the dense
                                   (final layer) which is a
                                   :py:class:`keras.layers.Dense` object. If no
                                   parameters to pass leave as None.
        :param optimiser: Optimiser to be used accepts any
                          `keras optimiser <https://keras.io/optimizers/>`_.
                          Default is :py:class:`keras.optimizers.SGD`
        :param optimiser_params: Parameters for the optimiser. If None uses
                                 default optimiser parameters.
        '''

        self.tokeniser = tokeniser
        self.embeddings = embeddings
        self.reproducible = reproducible
        self.pad_size = pad_size
        self.test_pad_size = 0
        self.lower = lower
        self.patience = patience
        self.batch_size = batch_size
        self.epochs = epochs

        self.optimiser_params = optimiser_params
        if optimiser_params is None:
            self.optimiser_params = {}

        self.embedding_layer_kwargs = embedding_layer_kwargs
        if embedding_layer_kwargs is None:
            self.embedding_layer_kwargs = {}
        self.lstm_layer_kwargs = lstm_layer_kwargs
        if lstm_layer_kwargs is None:
            self.lstm_layer_kwargs = {}
        self.dense_layer_kwargs = dense_layer_kwargs
        if dense_layer_kwargs is None:
            self.dense_layer_kwargs = {}
        self.optimiser = optimiser
        self.model = None
        self.fitted = False

    def model_parameters(self) -> Dict[str, Any]:
        '''
        Returns a dictionary containing the attributes of the class instance,
        the parameters to give to the class constructior to re-create this
        instance, and the class itself.

        This is used by the :py:meth:`save` method so that the instance can
        be re-created when loaded by the :py:meth:`load` method.
        '''

        class_params = {'tokeniser': self.tokeniser,
                        'embeddings': self.embeddings,
                        'reproducible': self.reproducible,
                        'pad_size': self.pad_size,
                        'lower': self.lower,
                        'patience': self.patience,
                        'batch_size': self.batch_size, 'epochs': self.epochs,
                        'embedding_layer_kwargs': self.embedding_layer_kwargs,
                        'lstm_layer_kwargs': self.lstm_layer_kwargs,
                        'dense_layer_kwargs': self.dense_layer_kwargs,
                        'optimiser': self.optimiser,
                        'optimiser_params': self.optimiser_params}
        class_attrs = {'test_pad_size': self.test_pad_size}
        return {'class': self.__class__, 'class_attrs': class_attrs,
                'class_params': class_params}

    def _pre_process(self, data_dicts: List[Dict[str, str]],
                     training: bool = False) -> np.ndarray:
        '''
        Converts the text in the data_dicts into a matrix of shape
        [n_samples, pad_size] where each integer in the matrix represents
        the word embedding lookup. This is then used as input into the
        keras model.

        The text from the data_dicts are converted by the
        :py:meth:`process_text` method.

        :param data_dicts: A list of dictonaries that contains a `text` field.
        :param training: Whether the text should be processed for training or
                         for prediction. prediction = False, training = True
        :return: The output of :py:meth:`process_text` method.
        '''
        text_data = [data['text'] for data in data_dicts]
        if training:
            pad_data = self.process_text(text_data, self.pad_size)
            self.test_pad_size, sequence_data = pad_data
            return sequence_data
        _, sequence_data = self.process_text(text_data, self.test_pad_size)
        return sequence_data

    def create_training_y(self, train_y: np.ndarray, validation_y: np.ndarray,
                          ) -> Tuple[np.ndarray, np.ndarray]:
        '''
        Converts the training and validation target values from a vector of
        class lables into a matrix of binary values of shape [n_samples,
        n_classes].

        To convert the vector of classes to a matrix we the
        :py:func:`keras.utils.to_categorical` function.

        :param train_y: Vector of class labels, shape = [n_samples]
        :param validation_y: Vector of class labels, shape = [n_samples]
        :return: A tuple of length two containing the train and validation
                 matrices respectively. The shape of each matrix is:
                 [n_samples, n_classes]
        '''
        train_y = to_categorical(train_y).astype(np.float32)
        validation_y = to_categorical(validation_y).astype(np.float32)
        return train_y, validation_y

    def create_training_text(self, train_data: List[Dict[str, str]],
                             validation_data: List[Dict[str, str]]
                             ) -> Tuple[np.ndarray, np.ndarray]:
        '''
        Converts the training and validation data into a format that the keras
        model can take as input.

        :param train_data: Data to be trained on. Which is a list of
                           dictionaries where each dictionary has a `text`
                           field containing text.
        :param validation_data: Data to evaluate the model at training time.
                                Which is a list of dictionaries where each
                                dictionary has a `text` field containing text.
        :return: A tuple of length two containing the train and validation
                 input that are both the output of :py:meth:`_pre_process`
        '''
        train_sequence = self._pre_process(train_data, training=True)
        val_sequence = self._pre_process(validation_data, training=False)
        return train_sequence, val_sequence

    def keras_model(self, num_classes: int) -> 'keras.models.Model':
        '''
        The model that represents this class. This is a single forward LSTM.

        :param num_classes: Number of classes to predict.
        :return: Forward LSTM keras model.
        '''
        # Embeddings
        embedding_matrix = self.embeddings.embedding_matrix
        vocab_size, vector_size = embedding_matrix.shape

        embedding_layer_kwargs = self.embedding_layer_kwargs
        embedding_layer_trainable = True
        if 'trainable' in embedding_layer_kwargs:
            embedding_layer_trainable = embedding_layer_kwargs.pop('trainable')

        lstm_layer_kwargs = self.lstm_layer_kwargs
        lstm_dimension = vector_size
        if 'cell' in self.lstm_layer_kwargs:
            lstm_dimension = lstm_layer_kwargs.pop('cell')

        dense_layer_kwargs = self.dense_layer_kwargs
        # output_activation = 'softmax' if num_classes > 2 else ''
        # Model layers
        input_layer = layers.Input(shape=(self.test_pad_size,),
                                   name='text_input')
        embedding_layer = layers\
                          .Embedding(input_dim=vocab_size,
                                     output_dim=vector_size,
                                     input_length=self.test_pad_size,
                                     trainable=embedding_layer_trainable,
                                     weights=[embedding_matrix],
                                     name='embedding_layer',
                                     **embedding_layer_kwargs
                                     )(input_layer)
        lstm_layer = layers.LSTM(lstm_dimension,
                                 name='lstm_layer',
                                 **lstm_layer_kwargs)(embedding_layer)
        prediction_layer = layers.Dense(num_classes, activation='softmax',
                                        name='output',
                                        **dense_layer_kwargs)(lstm_layer)

        return models.Model(inputs=input_layer, outputs=prediction_layer)

    @property
    def pad_size(self) -> int:
        '''
        pad_size attribute

        :return: The pad_size used in the model
        '''

        return self._pad_size

    @pad_size.setter
    def pad_size(self, value: int) -> None:
        '''
        Sets the pad_size attribute

        :param value: The value to assign to the pad_size attribute
        '''

        self.fitted = False
        self._pad_size = value

    @property
    def embedding_layer_kwargs(self) -> Dict[str, Any]:
        '''
        embedding_layer_kwargs attribute

        :return: The embedding_layer_kwargs used in the model
        '''

        return self._embedding_layer_kwargs

    @embedding_layer_kwargs.setter
    def embedding_layer_kwargs(self, value: Dict[str, Any]) -> None:
        '''
        Sets the embedding_layer_kwargs attribute

        :param value: The value to assign to the embedding_layer_kwargs
                      attribute
        '''

        self.fitted = False
        self._embedding_layer_kwargs = value

    @property
    def lstm_layer_kwargs(self) -> Dict[str, Any]:
        '''
        lstm_layer_kwargs attribute

        :return: The lstm_layer_kwargs used in the model
        '''

        return self._lstm_layer_kwargs

    @lstm_layer_kwargs.setter
    def lstm_layer_kwargs(self, value: Dict[str, Any]) -> None:
        '''
        Sets the lstm_layer_kwargs attribute

        :param value: The value to assign to the lstm_layer_kwargs
                      attribute
        '''

        self.fitted = False
        self._lstm_layer_kwargs = value

    @property
    def dense_layer_kwargs(self) -> Dict[str, Any]:
        '''
        dense_layer_kwargs attribute

        :return: The dense_layer_kwargs used in the model
        '''

        return self._dense_layer_kwargs

    @dense_layer_kwargs.setter
    def dense_layer_kwargs(self, value: Dict[str, Any]) -> None:
        '''
        Sets the dense_layer_kwargs attribute

        :param value: The value to assign to the dense_layer_kwargs
                      attribute
        '''

        self.fitted = False
        self._dense_layer_kwargs = value


class TDLSTM(LSTM):
    '''
    Attributes:

    1. include_target -- Wheather to include the target in the LSTM
       representations.
    '''

    @classmethod
    def name(cls) -> str:
        return 'TDLSTM'

    def __repr__(self) -> str:
        '''
        Name of the machine learning model.
        '''
        return self.name()

    def __init__(self, tokeniser: Callable[[str], List[str]],
                 embeddings: 'bella.word_vectors.WordVectors',
                 reproducible: Union[int, None] = None, pad_size: int = -1,
                 lower: bool = True, patience: int = 10,
                 batch_size: int = 32, epochs: int = 300,
                 embedding_layer_kwargs: Dict[str, Any] = None,
                 lstm_layer_kwargs: Dict[str, Any] = None,
                 dense_layer_kwargs: Dict[str, Any] = None,
                 optimiser: 'keras.optimizers.Optimizer' = optimizers.SGD,
                 optimiser_params: Union[Dict[str, Any], None] = None,
                 include_target: bool = True) -> None:
        '''
        :param tokeniser: Tokeniser to be used e.g. :py:meth:`str.split`.
        :param embeddings: Embedding (Word vectors) to be used e.g.
                           :py:class:`bella.word_vectors.SSWE`
        :param reproducible: Whether to be reproducible. If None then it is
                             but quicker to run. Else provide a `int` that
                             will represent the random seed value.
        :param pad_size: The max number of tokens to use per sequence. If -1
                         use the text sequence in the training data that has
                         the most tokens as the pad size.
        :param lower: Whether to lower case the words being processed.
        :param patience: Number of epochs with no improvement before training
                         is stopped.
        :param batch_size: Number of samples per gradient update.
        :param epochs: Number of times to train over the entire training set
                       before stopping. If patience is set, then it may
                       stop before reaching the number of epochs specified
                       here.
        :param embedding_layer_kwargs: Keyword arguments to pass to the
                                       embedding layer which is a
                                       :py:class:`keras.layers.Embedding`
                                       object. If no parameters to pass leave
                                       as None.
        :param lstm_layer_kwargs: Keyword arguments to pass to the lstm
                                  layer(s) which is a
                                  :py:class:`keras.layers.LSTM` object. If no
                                  parameters to pass leave as None.
        :param dense_layer_kwargs: Keyword arguments to pass to the dense
                                   (final layer) which is a
                                   :py:class:`keras.layers.Dense` object. If no
                                   parameters to pass leave as None.
        :param optimiser: Optimiser to be used accepts any
                          `keras optimiser <https://keras.io/optimizers/>`_.
                          Default is :py:class:`keras.optimizers.SGD`
        :param optimiser_params: Parameters for the optimiser. If None uses
                                 default optimiser parameters.
        :param include_target: Wheather to include the target in the LSTM
                               representations.
        '''

        super().__init__(tokeniser, embeddings, reproducible, pad_size, lower,
                         patience, batch_size, epochs, embedding_layer_kwargs,
                         lstm_layer_kwargs, dense_layer_kwargs, optimiser,
                         optimiser_params)

        self.left_pad_size = pad_size
        self.left_test_pad_size = 0
        self.right_pad_size = pad_size
        self.right_test_pad_size = 0
        self.include_target = include_target

    def model_parameters(self) -> Dict[str, Any]:
        '''
        Returns a dictionary containing the attributes of the class instance,
        the parameters to give to the class constructior to re-create this
        instance, and the class itself.

        This is used by the :py:meth:`save` method so that the instance can
        be re-created when loaded by the :py:meth:`load` method.
        '''

        attributes = super().model_parameters()
        class_attrs = {'left_test_pad_size': self.left_test_pad_size,
                       'right_test_pad_size': self.right_test_pad_size}
        attributes['class_attrs'] = class_attrs

        class_params = attributes['class_params']
        class_params['include_target'] = self.include_target
        attributes['class_params'] = class_params
        return attributes

    def _pre_process(self, data_dicts: List[Dict[str, Any]],
                     training: bool = False) -> List[np.ndarray]:
        '''
        Converts the text in the data_dicts into a List of size two
        representing the left and right context of the target word
        respectively. Each List is made up of a matrix of of integers
        representing the text as their embedding lookups. These two Lists
        are the inputs into the keras model.

        Two find the left and right contexts it uses the `spans` field of
        the dictionaries in the `data_dicts`. The `spans` field is a list of
        Tuples where each Tuple represents a occurence of the Target, each
        Tuple contains the index of the starting and ending character index
        (Expects the List to be of size 1 as there should be only one target
        per target sample. This case is not True for the
        `Dong et al. <https://aclanthology.info/papers/P14-2009/p14-2009>`_
        dataset therefore it only takes the first target instance in the
        sentence as the target).

        The texts are converted into integers using the
        :py:meth:`process_text` method.

        :param data_dicts: A list of dictonaries that contains a `text` and
                           `spans` field.
        :param training: Whether the text should be processed for training or
                         for prediction. prediction = False, training = True
        :return: A list of two contaning the left and right context of
                 the target both represented by the output of
                 :py:meth:`process_text` method.
        '''

        def context_texts(context_data_dicts: List[Dict[str, Any]]
                          ) -> Tuple[List[str], List[str]]:
            '''
            :param context_data_dicts: A list of dictonaries that contains a
                                       `text` and `spans` field.
            :return: A list of the left and right text contexts for all the
                     dictionaries.
            '''
            # Context returns all of the left and right context occurrences
            # therefore if a target is mentioned Twice and are associated then
            # for a single text two left and right occurrences are returned.
            # Thus these are a list of lists we therefore chose only the
            # first mentioned target as the paper linked to this method does
            # not specify which they used.
            left_texts = [context(data, 'left', inc_target=self.include_target)
                          for data in context_data_dicts]
            right_texts = [context(data, 'right',
                                   inc_target=self.include_target)
                           for data in context_data_dicts]
            left_texts = [texts[0] for texts in left_texts]
            right_texts = [texts[0] for texts in right_texts]
            return left_texts, right_texts

        # Convert from a sequence of dictionaries into texts and then integers
        # that represent the tokens in the text within the embedding space.

        # Get left and right contexts
        left_text, right_text = context_texts(data_dicts)
        if training:
            left_pad_sequence = self.process_text(left_text,
                                                  self.left_pad_size)
            self.left_test_pad_size, left_sequence = left_pad_sequence

            right_pad_sequence = self.process_text(right_text,
                                                   self.right_pad_size,
                                                   padding='post',
                                                   truncate='post')
            self.right_test_pad_size, right_sequence = right_pad_sequence
            return left_sequence, right_sequence

        left_pad_sequence = self.process_text(left_text,
                                              self.left_test_pad_size)
        _, left_sequence = left_pad_sequence

        right_pad_sequence = self.process_text(right_text,
                                               self.right_test_pad_size,
                                               padding='post',
                                               truncate='post')
        _, right_sequence = right_pad_sequence
        return [left_sequence, right_sequence]

    def create_training_text(self, train_data: List[Dict[str, Any]],
                             validation_data: List[Dict[str, Any]]
                             ) -> Tuple[List[np.ndarray],
                                        List[np.ndarray]]:
        '''
        Converts the training and validation data into a format that the keras
        model can take as input.

        :param train_data: Data to be trained on. Which is a list of
                           dictionaries where each dictionary has a `text`
                           field containing text and a field `spans` containing
                           a list of Tuples where each Tuple represents a
                           occurence of the Target, each Tuple contains the
                           index of the starting and ending character index
                           (Expects the List to be of size 1 as there should
                           be only one target per target sample. This case is
                           not True for the
                           `Dong et al. <https://aclanthology.info/papers/P14-\
                           2009/p14-2009>`_ dataset therefore it only takes
                           the first target instance in the sentence as the
                           target).
        :param validation_data: Data to evaluate the model at training time.
                                Expects the same data as the `train_data`
                                parameter.
        :return: A tuple of length two containing the train and validation
                 input that are both the output of :py:meth:`_pre_process`
        '''

        return super().create_training_text(train_data, validation_data)

    def keras_model(self, num_classes: int) -> 'keras.models.Model':
        '''
        The model that represents this class. This is a custom combination
        of two LSTMs.

        :param num_classes: Number of classes to predict.
        :return: Two LSTMs, one forward from the left context and the other
                 backward from the right context. The output of the two are
                 concatenated and are input to the output layer.
        '''
        # Embeddings
        embedding_matrix = self.embeddings.embedding_matrix
        vocab_size, vector_size = embedding_matrix.shape

        embedding_layer_kwargs = self.embedding_layer_kwargs
        embedding_layer_trainable = True
        if 'trainable' in embedding_layer_kwargs:
            embedding_layer_trainable = embedding_layer_kwargs.pop('trainable')

        lstm_layer_kwargs = self.lstm_layer_kwargs
        lstm_dimension = vector_size
        if 'cell' in self.lstm_layer_kwargs:
            lstm_dimension = lstm_layer_kwargs.pop('cell')

        dense_layer_kwargs = self.dense_layer_kwargs
        # Model layers
        # Left LSTM
        left_input = layers.Input(shape=(self.left_test_pad_size,),
                                  name='left_text_input')
        left_embedding_layer = layers\
                               .Embedding(input_dim=vocab_size,
                                          output_dim=vector_size,
                                          input_length=self.left_test_pad_size,
                                          trainable=embedding_layer_trainable,
                                          weights=[embedding_matrix],
                                          name='left_embedding_layer',
                                          **embedding_layer_kwargs
                                          )(left_input)
        left_lstm_layer = layers.LSTM(lstm_dimension,
                                      name='left_lstm_layer',
                                      **lstm_layer_kwargs
                                      )(left_embedding_layer)
        # Right LSTM
        right_input = layers.Input(shape=(self.right_test_pad_size,),
                                   name='right_text_input')
        right_embedding_layer = layers\
                                .Embedding(input_dim=vocab_size,
                                           output_dim=vector_size,
                                           input_length=self.right_test_pad_size,
                                           trainable=embedding_layer_trainable,
                                           weights=[embedding_matrix],
                                           name='right_embedding_layer',
                                           **embedding_layer_kwargs
                                           )(right_input)
        right_lstm_layer = layers.LSTM(lstm_dimension,
                                       name='right_lstm_layer',
                                       go_backwards=True,
                                       **lstm_layer_kwargs
                                       )(right_embedding_layer)
        # Merge the outputs of the left and right LSTMs
        merge_layer = layers.concatenate([left_lstm_layer, right_lstm_layer],
                                         name='left_right_lstm_merge')
        predictions = layers.Dense(num_classes, activation='softmax',
                                   name='output',
                                   **dense_layer_kwargs)(merge_layer)

        return models.Model(inputs=[left_input, right_input],
                            outputs=predictions)

    @property
    def include_target(self) -> bool:
        '''
        include_target attribute

        :return: The include_target used in the model
        '''

        return self._include_target

    @include_target.setter
    def include_target(self, value: bool) -> None:
        '''
        Sets the include_target attribute

        :param value: The value to assign to the include_target attribute
        '''

        self.fitted = False
        self._include_target = value


class TCLSTM(TDLSTM):

    @classmethod
    def name(cls) -> str:
        return 'TCLSTM'

    def __repr__(self) -> str:
        '''
        Name of the machine learning model.
        '''
        return self.name()

    def _pre_process(self, data_dicts: List[Dict[str, Any]],
                     training: bool = False) -> List[np.ndarray]:
        '''
        Converts the text in the data_dicts into a list of size four
        representing the left context, left targets, right context and
        right targets. Where the contexts come are the same as those from
        TDLSTM :py:meth:`bella.models.tdlstm.TDLSTM._pre_process` method.

        The targets are a matrix of size [word_embedding_dimension, pad_size]
        and each vector in the matrix is the word embedding representation
        of the target word. If the target word is made up of multiple words
        it is then the average of the words vector representation (we use the
        median as the average). Both the contexts and the target matrix are
        used as input into the keras model.

        The texts are converted into integers using the
        :py:meth:`process_text` method.

        :param data_dicts: A list of dictonaries that contains a `text` and
                           `spans` field.
        :param training: Whether the text should be processed for training or
                         for prediction. prediction = False, training = True
        :return: A list of four contaning the left context, left vectors,
                 right context, and right vectors.
        '''
        def context_median_targets(pad_size: int):
            '''
            :param pad_size: The number of timesteps within the LSTM
            :return: Matrix of size [word_embedding_dimension, pad_size] where
                     each word embedding represents the target word or if
                     multiple words make up the target the word embedding is
                     the median of the words embeddings.
            '''
            vector_size = self.embeddings.vector_size
            target_matrix = np.zeros((len(data_dicts),
                                      pad_size, vector_size))
            for index, data in enumerate(data_dicts):
                target_vectors = []
                target_words = data['target'].split()
                for target_word in target_words:
                    if self.lower:
                        target_word = target_word.lower()
                    target_embedding = self.embeddings\
                                           .lookup_vector(target_word)
                    target_vectors.append(target_embedding)
                target_vectors = np.vstack(target_vectors)
                median_target_vector = matrix_median(target_vectors)
                median_vectors = np.repeat(median_target_vector, pad_size,
                                           axis=0)
                target_matrix[index] = median_vectors
            return target_matrix

        sequences = super()._pre_process(data_dicts, training=training)
        left_sequence, right_sequence = sequences
        left_target_vectors = context_median_targets(self.left_test_pad_size)
        right_target_vectors = context_median_targets(self.right_test_pad_size)
        return [left_sequence, left_target_vectors,
                right_sequence, right_target_vectors]

    def create_training_text(self, train_data: List[Dict[str, Any]],
                             validation_data: List[Dict[str, Any]]
                             ) -> Tuple[List[np.ndarray],
                                        List[np.ndarray]]:
        '''
        Converts the training and validation data into a format that the keras
        model can take as input.

        :param train_data: See :py:meth:`bella.models.tdlstm.\
                           TDLSTM.create_training_text` `train_data`
                           parameter.
        :param validation_data: See :py:meth:`bella.models.tdlstm.\
                                TDLSTM.create_training_text` `validation_data`
                                parameter.
        :return: A tuple of length two containing the train and validation
                 input that are both the output of :py:meth:`_pre_process`
        '''

        return super().create_training_text(train_data, validation_data)

    def keras_model(self, num_classes: int) -> 'keras.models.Model':
        '''
        The model that represents this class. This is the same as the
        :py:meth:`bella.models.tdlstm.TDLSTM.keras_model` model, however
        the words in before inputting into the LSTM are concatenated with
        the word embedding of the target. If the target is more than one word
        then the word embedding of the target is the average (median in our
        case) embeddings of the target words.

        :param num_classes: Number of classes to predict.
        :return: Two LSTMs one forward from the left context and the other
                 backward from the right context taking into account the
                 target vector embedding.
        '''
        # Embeddings
        embedding_matrix = self.embeddings.embedding_matrix
        vocab_size, vector_size = embedding_matrix.shape

        embedding_layer_kwargs = self.embedding_layer_kwargs
        embedding_layer_trainable = True
        if 'trainable' in embedding_layer_kwargs:
            embedding_layer_trainable = embedding_layer_kwargs.pop('trainable')

        lstm_layer_kwargs = self.lstm_layer_kwargs
        # Double the vector size as we have to take into consideration the
        # concatenated target vector
        lstm_dimension = vector_size * 2
        if 'cell' in self.lstm_layer_kwargs:
            lstm_dimension = lstm_layer_kwargs.pop('cell')

        dense_layer_kwargs = self.dense_layer_kwargs
        # Model layers
        # Left LSTM
        left_input = layers.Input(shape=(self.left_test_pad_size,),
                                  name='left_text_input')
        left_embedding_layer = layers\
                               .Embedding(input_dim=vocab_size,
                                          output_dim=vector_size,
                                          input_length=self.left_test_pad_size,
                                          trainable=embedding_layer_trainable,
                                          weights=[embedding_matrix],
                                          name='left_embedding_layer',
                                          **embedding_layer_kwargs
                                          )(left_input)
        left_target_input = layers.Input(shape=(self.left_test_pad_size,
                                                vector_size),
                                         name='left_target')
        left_text_target = layers.concatenate([left_embedding_layer,
                                               left_target_input],
                                              name='left_text_target')
        left_lstm_layer = layers.LSTM(lstm_dimension,
                                      name='left_lstm_layer',
                                      **lstm_layer_kwargs
                                      )(left_text_target)
        # Right LSTM
        right_input = layers.Input(shape=(self.right_test_pad_size,),
                                   name='right_text_input')
        right_embedding_layer = layers\
                                .Embedding(input_dim=vocab_size,
                                           output_dim=vector_size,
                                           input_length=self.right_test_pad_size,
                                           trainable=embedding_layer_trainable,
                                           weights=[embedding_matrix],
                                           name='right_embedding_layer',
                                           **embedding_layer_kwargs
                                           )(right_input)
        right_target_input = layers.Input(shape=(self.right_test_pad_size,
                                                 vector_size),
                                          name='right_target')
        right_text_target = layers.concatenate([right_embedding_layer,
                                                right_target_input],
                                               name='right_text_target')
        right_lstm_layer = layers.LSTM(lstm_dimension,
                                       name='right_lstm_layer',
                                       go_backwards=True,
                                       **lstm_layer_kwargs
                                       )(right_text_target)
        # Merge the outputs of the left and right LSTMs
        merge_layer = layers.concatenate([left_lstm_layer, right_lstm_layer],
                                         name='left_right_lstm_merge')
        predictions = layers.Dense(num_classes, activation='softmax',
                                   name='output',
                                   **dense_layer_kwargs)(merge_layer)

        input_layers = [left_input, left_target_input,
                        right_input, right_target_input]
        return models.Model(inputs=input_layers, outputs=predictions)



'''
Module that contains the various data types:

1. Target -- Mutable data store for a single Target value i.e. one training \
example.
2. TargetCollection -- Mutable data store for Target data types.  i.e. A \
data store that contains multiple Target instances.
'''

from collections.abc import MutableMapping
from collections import OrderedDict, defaultdict
import copy
from typing import List, Callable

import numpy as np
import pandas as pd
from sklearn import metrics
import seaborn as sns

from bella.tokenisers import whitespace
from bella.stanford_tools import constituency_parse

class Target(MutableMapping):
    '''
    Mutable data store for a single Target value. This should be used as the
    value for Target information where it contains all data required to be
    classified as Target data for Target based sentiment classification.

    Overrides collections.abs.MutableMapping abstract class.

    Reference on how I created the class:
    http://www.kr41.net/2016/03-23-dont_inherit_python_builtin_dict_type.html
    https://docs.python.org/3/library/collections.abc.html

    Functions changed compared to normal:

    1. __delitem__ -- Will only delete the `target_id` key.
    2. __eq__ -- Two Targets are the same if they either have the same `id` or \
    they have the same values to the minimum keys \
    ['spans', 'text', 'target', 'sentiment']
    3. __setitem__ -- Only allows you to add/modify the `predicted` key which \
    represents the predicted sentiment for the Target instance.
    '''

    def __init__(self, spans, target_id, target, text, sentiment, predicted=None,
                 sentence_id=None):
        '''
        :param target: Target that the sentiment is about. e.g. Iphone
        :param sentiment: sentiment of the target.
        :param text: The text context that the target and sentiment is within.
        :param target_id: Unique ID. Has to be Unique within the \
        TargetCollection if it is put into a TargetCollection.
        :param spans: List of tuples where each tuple is of length 2 where they \
        contain the exclusive range of an instance of Target word in the Text \
        context. The reason it is a list if because the Target word can be \
        mentioned more than once e.g. `The Iphone was great but the iphone is \
        small`. The first Int in the tuple has to be less than the second Int.
        :param predicted: If given adds the predicted sentiment value.
        :param sentence_id: Unique ID of the sentence that the target is \
        within. More than one target can have the same sentence.
        :type target: String
        :type sentiment: String or Int (Based on annotation schema)
        :type text: String
        :type target_id: String
        :type spans: list
        :type predicted: Same type as sentiment. Default None (Optional)
        :type sentence_id: String. Default None (Optional)
        :returns: Nothing. Constructor.
        :rtype: None
        '''
        if not isinstance(target_id, str):
            raise TypeError('The target ID has to be of type String and not {}'\
                            .format(type(target_id)))
        if not isinstance(target, str):
            raise TypeError('The target has to be of type String and not {}'\
                            .format(type(target)))
        if not isinstance(text, str):
            raise TypeError('The text has to be of type String and not {}'\
                            .format(type(text)))
        if not isinstance(sentiment, (str, int)):
            raise TypeError('The sentiment has to be of type String or Int and '\
                            'not {}'.format(type(sentiment)))
        if not isinstance(spans, list):
            raise TypeError('The spans has to be of type list and not {}'\
                            .format(type(spans)))
        else:
            if len(spans) < 1:
                raise TypeError('spans has to contain at least one tuple not '\
                                'None')
            else:
                for span in spans:
                    if not isinstance(span, tuple):
                        raise TypeError('Spans has to be a list of tuples not {}'\
                                        .format(type(span)))
                    if len(span) != 2:
                        raise ValueError('Spans must contain tuples of length'\
                                         ' 2 not {}'.format(spans))
                    if not isinstance(span[0], int) or \
                       not isinstance(span[1], int):
                        raise TypeError('spans must be made of tuple containing '\
                                        'two Integers not {}'.format(span))
                    if span[1] <= span[0]:
                        raise ValueError('The first integer in a span must be '\
                                         'less than the second integer {}'\
                                         .format(span))
        temp_dict = dict(spans=spans, target_id=target_id, target=target,
                         text=text, sentiment=sentiment)
        if sentence_id is not None:
            if not isinstance(sentence_id, str):
                raise TypeError('`sentence_id` has to be a String and not {}'\
                                .format(type(sentence_id)))
            temp_dict['sentence_id'] = sentence_id

        self._storage = temp_dict
        if predicted is not None:
            self['predicted'] = predicted


    def __getitem__(self, key):
        return self._storage[key]

    def __iter__(self):
        return iter(self._storage)

    def __len__(self):
        return len(self._storage)

    def __delitem__(self, key):
        '''
        To ensure that the Target class maintains the minimum Keys and Values
        to allow an instance to be used in Target based machine learning. The
        key and associated values that can be deleted are limited to:

        1. target_id

        :param key: The key and associated value to delete from the store.
        :returns: Updates the data store by removing key and value.
        :rtype: None
        '''

        accepted_keys = set(['target_id'])
        if key not in accepted_keys:
            raise KeyError('The only keys that can be deleted are the '\
                           'following: {} the key you wish to delete {}'\
                           .format(accepted_keys, key))
        del self._storage[key]

    def __setitem__(self, key, value):
        '''
        :param key: key (Only store values for `predicted` key)
        :param value: Predicted sentiment value which has to be the same data \
        type as the `sentiment` value.
        :type key: String (`predicted` is the only key accepted at the moment)
        :type value: Int or String.
        :returns: Nothing. Adds the predicted sentiment of the Target.
        :rtype: None.
        '''

        if key != 'predicted':
            raise KeyError('The Only key that can be changed is the `predicted`'\
                           ' key not {}'.format(key))
        raise_type = False
        sent_value = self._storage['sentiment']
        if isinstance(sent_value, int):
            if not isinstance(value, (int, np.int32, np.int64)):
                raise_type = True
        elif not isinstance(value, type(sent_value)):
            raise_type = True

        if raise_type:
            raise TypeError('Value to be stored for the `predicted` sentiment '\
                            'has to be the same data type as the sentiment '\
                            'value {} and not {}.'\
                            .format(sent_value, type(value)))
        self._storage[key] = value

    def __repr__(self):
        '''
        :returns: String returned is what user see when the instance is \
        printed or printed within a interpreter.
        :rtype: String
        '''

        return 'Target({})'.format(self._storage)

    def __eq__(self, other):
        '''
        Two Target instances are equal if they are both Target instances and
        one of the following conditions

        1. They have the same target_id (This is preferred)
        2. The minimum keys that all targets have to have \
        ['spans', 'text', 'target', 'sentiment'] are all equal.

        :param other: The target instance that is being compare to the current \
        target instance.
        :type other: Target
        :returns: True if they are equal else False.
        :rtype: bool
        '''

        if not isinstance(other, Target):
            return False
        if 'target_id' in self and 'target_id' in other:
            if self['target_id'] != other['target_id']:
                return False
        else:
            minimum_keys = ['spans', 'text', 'target', 'sentiment']
            for key in minimum_keys:
                if not self[key] == other[key]:
                    return False
        return True
    def __array__(self):
        '''
        Function for converting it to a numpy array
        '''
        return np.asarray(dict(self))

class TargetCollection(MutableMapping):
    '''
    Mutable data store for Target data types.  i.e. A data store that contains
    multiple Target instances. This collection ensures that there are no two
    Target instances stored that have the same ID this is because the storage
    of the collection is an OrderedDict.

    Overrides collections.abs.MutableMapping abstract class.

    Functions:

    1. add -- Given a Target instance with an `id` key adds it to the data \
    store.
    2. data -- Returns all of the Target instances stored as a list of Target \
    instances.
    3. stored_sentiments -- Returns a set of unique sentiments stored.
    4. sentiment_data -- Returns the list of all sentiment values stored in \
    the Target instances stored.
    5. add_pred_sentiment -- Adds a list of predicted sentiment values to the \
    Target instances stored.
    6. confusion_matrix -- Returns the confusion matrix between the True and \
    predicted sentiment values.
    7. subset_by_sentiment -- Creates a new TargetCollection based on the \
    number of unique sentiments in a sentence.
    '''

    def __init__(self, list_of_target=None):
        '''
        :param list_of_target: An interator of Target instances e.g. a List of \
        Target instances.
        :type list_of_target: Iterable. Default None (Optional)
        :returns: Nothing. Constructor.
        :rtype: None
        '''

        self._storage = OrderedDict()
        if list_of_target is not None:
            if not hasattr(list_of_target, '__iter__'):
                raise TypeError('The list_of_target argument has to be iterable')
            for target in list_of_target:
                self.add(target)

    def __getitem__(self, key):
        return self._storage[key]

    def __setitem__(self, key, value):
        '''
        If key already exists will raise KeyError.

        :param key: key that stores the index to the value
        :param value: value to store at the keys location
        :type key: hashable object
        :type value: Target
        :returns: Nothing. Adds data to the collection
        :rtype: None.
        '''

        if not isinstance(value, Target):
            raise TypeError('All values in this store have to be of type '\
                            'Target not {}'.format(type(value)))
        if key in self._storage:
            raise KeyError('This key: `{}` already exists with value `{}` '\
                           'value that for the same key is `{}`'\
                           .format(key, self._storage[key], value))
        temp_value = copy.deepcopy(value)
        # As the id will be saved as the key no longer needed in the target
        # instance (value). However if the key does not match the `target_id`
        # raise KeyError
        if 'target_id' in value:
            if value['target_id'] != key:
                raise KeyError('Cannot add this to the data store as the key {}'\
                               ' is not the same as the `target_id` in the Target'\
                               ' instance value {}'.format(key, value))
            del temp_value['target_id']
        self._storage[key] = temp_value

    def __delitem__(self, key):
        del self._storage[key]

    def __iter__(self):
        return iter(self._storage)

    def __len__(self):
        return len(self._storage)

    def add(self, value):
        '''
        Adds the Target instance to the data store without having to extract
        out the target_id of the target if using __setitem__

        :Example:

        >>> target = Target([(10, 16)], '1', 'Iphone',
                            'text with Iphone', 0)
        >>> target_col = TargetCollection()
        # Add method is simpler to use than __setitem__
        >>> target_col.add(target)
        # Example of the __setitem__ method
        >>> target_col[target['target_id']] = target

        :param value: Target instance with a `target_id` key
        :type value: Target
        :returns: Nothing. Adds the target instance to the data store.
        :rtype: None
        '''

        if not isinstance(value, Target):
            raise TypeError('All values in this store have to be of type '\
                            'Target not {}'.format(type(value)))
        if 'target_id' not in value:
            raise ValueError('The Target instance given {} does not have a '\
                             'target_id'.format(value))
        self[value['target_id']] = value

    def data(self):
        '''
        :returns: a list of all the Target instances stored.
        :rtype: list
        '''

        _data = []
        for _id, target_data in self.items():
            data_dict = {**target_data}
            data_dict['target_id'] = _id
            _data.append(Target(**data_dict))
        return _data

    def data_dict(self):
        '''
        :returns: Same as the data function but returns dicts instead of \
        Targets
        :rtype: list
        '''

        _data = []
        for _id, target_data in self.items():
            data_dict = {**target_data}
            data_dict['target_id'] = _id
            _data.append(data_dict)
        return _data

    def stored_sentiments(self):
        '''
        :returns: A set of all unique sentiment values of the target instances \
        in the data store.
        :rtype: set
        '''

        unique_sentiments = set()
        for target_data in self.values():
            unique_sentiments.add(target_data['sentiment'])
        return unique_sentiments

    def sentiment_data(self, mapper=None, sentiment_field='sentiment'):
        '''
        :param mapper: A dictionary that maps the keys to the values where the \
        keys are the current unique sentiment values of the target instances \
        stored
        :param sentiment_field: Determines if it should return True sentiment \
        of the Targets `sentiment` or to return the predicted value `predicted`
        :type mapper: dict
        :type sentiment_field: String. Default `sentiment` (True values)
        :returns: a list of the sentiment value for each Target instance stored.
        :rtype: list

        :Example of using the mapper:
        >>> target_col = TargetCollection([Target([(10, 16)], '1', 'Iphone',
                                                  'text with Iphone', 'pos')])
        >>> target_col.add(Target([(10, 15)], '2', 'Pixel',
                                  'text with Pixel', 'neg'))
        # Get the unique sentiment values for each target instance
        >>> map_keys = target_col.stored_sentiments()
        >>> print(map_keys)
        >>> ['pos', 'neg']
        >>> mapper = {'pos' : 1, 'neg' : -1}
        >>> print(target_col.sentiment_data(mapper=mapper))
        >>> 1, -1
        '''

        allowed_fields = set(['sentiment', 'predicted'])
        if sentiment_field not in allowed_fields:
            raise ValueError('The `sentiment_field` has to be one of the '\
                             'following values {} and not {}'\
                             .format(allowed_fields, sentiment_field))

        if mapper is not None:
            if not isinstance(mapper, dict):
                raise TypeError('The mapper has to be of type dict and not {}'\
                                .format(type(mapper)))
            allowed_keys = self.stored_sentiments()
            if len(mapper) != len(allowed_keys):
                raise ValueError('The mapper has to contain a mapping for each '\
                                 'unique sentiment value {} and not a subset '\
                                 'given {}'.format(allowed_keys, mapper.keys()))
            for map_key in mapper:
                if map_key not in allowed_keys:
                    raise ValueError('The mappings are not correct. The map '\
                                     'key {} does not exist in the unique '\
                                     'sentiment values in the store {}'\
                                     .format(map_key, allowed_keys))
            return [mapper[target_data[sentiment_field]]\
                    for target_data in self.values()]

        return [target_data[sentiment_field] for target_data in self.values()]

    def add_id_pred(self, id_pred):
        count = 0
        for targ_id in self:
            if targ_id in id_pred:
                self[targ_id]['predicted'] = id_pred[targ_id]
                count += 1
        if count != len(self):
            raise ValueError('We have only added {} predictions to {} targets'\
                             .format(count, len(self)))

    def add_pred_sentiment(self, sent_preds, mapper=None):
        '''
        :param sent_preds: A list of predicted sentiments for all Target \
        instances stored.
        :param mapper: A dictionary mapping the predicted sentiment to \
        alternative values e.g. Integer values to String values.
        :type sent_preds: list
        :type mapper: dict
        :returns: Nothing. Adds the predicted sentiments to the Target \
        instances stored.
        :rtype: None
        '''

        if not isinstance(sent_preds, list):
            raise TypeError('The predicted sentiments have to be of type list '\
                            'not {}'.format(type(sent_preds)))
        if len(sent_preds) != len(self):
            raise ValueError('The length of the predicted sentiments {} is not '\
                             'equal to the number Target instances stored {}'\
                             .format(len(sent_preds), len(self)))
        for index, target in enumerate(self.data()):
            predicted_sent = sent_preds[index]
            if mapper is not None:
                predicted_sent = mapper[predicted_sent]
            target_id = target['target_id']
            self._storage[target_id]['predicted'] = predicted_sent

    def confusion_matrix(self, plot=False, norm=False):
        '''
        :param plot: To return a heatmap of the confusion matrix.
        :param norm: Normalise the values in the confusion matrix
        :type plot: bool. Default False
        :type norm: bool. Default False
        :returns: A tuple of length two. 1. the confusion matrix \
        2. The plot of the confusion matrix if plot is True else \
        None.
        :rtype: tuple
        '''

        sentiment_values = sorted(self.stored_sentiments())
        true_values = self.sentiment_data()
        pred_values = self.sentiment_data(sentiment_field='predicted')
        conf_matrix = metrics.confusion_matrix(true_values, pred_values,
                                               labels=sentiment_values)
        if norm:
            conf_matrix = conf_matrix / conf_matrix.sum()
        conf_matrix = pd.DataFrame(conf_matrix, columns=sentiment_values,
                                   index=sentiment_values)
        ax = None
        if plot:
            if norm:
                ax = sns.heatmap(conf_matrix, annot=True, fmt='.2f')
            else:
                ax = sns.heatmap(conf_matrix, annot=True, fmt='d')
        return conf_matrix, ax

    def subset_by_sentiment(self, num_unique_sentiments):
        '''
        Creates a subset based on the number of unique sentiment values per
        sentence. E.g. if num_unique_sentiments = 2 then it will
        return all the Target instances where each target intance has at least
        two target instances per sentence and those targets can have only one
        of two sentiment values. This can be used to test how well a method
        can extract exact sentiment information for the associated target.

        NOTE: Requires that all Target instances stored contain a sentence_id.

        :param num_unique_sentiments: Integer specifying the number of unique \
        sentiments in the target instances per sentence.
        :type num_unique_sentiments: int
        :returns: A subset based on the number of unique sentiments per sentence.
        :rtype: TargetCollection
        '''



        all_relevent_targets = []
        for targets in self.group_by_sentence().values():
            target_col = TargetCollection(targets)
            if len(target_col.stored_sentiments()) == num_unique_sentiments:
                all_relevent_targets.extend(targets)
        return TargetCollection(all_relevent_targets)

    def subset_by_sentence_length(self, length_condition):

        all_relevent_targets = []
        for target in self.data():
            target_text = target['text']
            if length_condition(target_text):
                all_relevent_targets.append(target)
        return TargetCollection(all_relevent_targets)

    # Not tested
    def targets_per_sentence(self):
        '''
        :returns: Dictionary of number of targets as keys and values the number \
        of sentences that have that many targets per sentence.
        :rtype: dict

        :Example:
        If we have 5 sentences that contain 1 target each and 4 sentences that
        contain 3 targets each then it will return a dict like:
        {1 : 5, 3 : 4}
        '''

        targets_sentence = {}
        for targets in self.group_by_sentence().values():
            num_targets = len(targets)
            targets_sentence[num_targets] = targets_sentence.get(num_targets, 0) + 1
        return targets_sentence

    # Not tested
    def avg_targets_per_sentence(self):
        return len(self) / self.number_sentences()
    # Not tested
    def number_sentences(self):
        return len(self.group_by_sentence())
    # Not tested
    def number_unique_targets(self):
        target_count = {}
        for target_instance in self.values():
            target = target_instance['target']
            target_count[target] = target_count.get(target, 0) + 1
        return len(target_count)

    def no_targets_sentiment(self):
        sentiment_targets = {}
        for target in self.values():
            sentiment = target['sentiment']
            sentiment_targets[sentiment] = sentiment_targets.get(sentiment, 0) + 1
        return sentiment_targets

    def ratio_targets_sentiment(self):
        no_sentiment_target = self.no_targets_sentiment()
        total_targets = sum(no_sentiment_target.values())
        ratio_sentiment_targets = {}
        for sentiment, no_targets in no_sentiment_target.items():
            ratio_sentiment_targets[sentiment] = round(no_targets / total_targets, 2)
        return ratio_sentiment_targets

    def group_by_sentence(self):
        '''
        :returns: A dictionary of sentence_id as keys and a list of target \
        instances that have the same sentence_id as values.
        :rtype: defaultdict (default is list)
        '''

        sentence_targets = defaultdict(list)
        for target in self.data():
            if 'sentence_id' not in target:
                raise ValueError('A Target id instance {} does not have '\
                                 'a sentence_id which is required.'\
                                 .format(target))
            sentence_id = target['sentence_id']
            sentence_targets[sentence_id].append(target)
        return sentence_targets

    def avg_constituency_depth(self):
        avg_depths = []
        for data in self.values():
            sentence_trees = constituency_parse(data['text'])
            tree_depths = [tree.height() - 1 for tree in sentence_trees]
            avg_depth = sum(tree_depths) / len(sentence_trees)
            avg_depths.append(avg_depth)
        return sum(avg_depths) / len(avg_depths)

    def avg_sentence_length_per_target(self, tokeniser=whitespace):
        all_sentence_lengths = []
        for data in self.values():
            all_sentence_lengths.append(len(tokeniser(data['text'])))
        return sum(all_sentence_lengths) / len(all_sentence_lengths)

    def word_list(self, tokeniser: Callable[[str], List[str]],
                  min_df: int = 0) -> List[str]:
        '''
        :param tokeniser: Tokeniser function to tokenise the text \
        of each target/sample
        :param min_df: Optional. The minimum percentage of documents a \
        token must occur in.
        :return: A word list of all tokens that occur in this data collection \
        given min_df.
        '''

        token_df = defaultdict(lambda: 0)
        num_df = 0
        for target in self.values():
            num_df += 1
            tokens = tokeniser(target['text'])
            for token in tokens:
                token_df[token.lower()] += 1
        min_df_value = int((num_df / 100) * min_df)
        word_list = [token for token, df in token_df.items()
                     if df > min_df_value]
        return word_list

    @staticmethod
    def combine_collections(*args):
        all_targets = []
        for collections in args:
            all_targets.extend(collections.data())
        return TargetCollection(all_targets)

    def __eq__(self, other):

        if len(self) != len(other):
            return False
        for key in self:
            if key not in other:
                return False
        return True

    def __repr__(self):
        '''
        :returns: String returned is what user see when the instance is \
        printed or printed within a interpreter.
        :rtype: String
        '''

        target_strings = ''

        self_len = len(self)
        if self_len > 2:
            for index, target in enumerate(self.data()):
                if index == 0:
                    target_strings += '{} ... '.format(target)
                if index == self_len - 1:
                    target_strings += '{}'.format(target)
        else:
            for target in self.data():
                target_strings += '{}, '.format(target)
        if target_strings != '':
            target_strings = target_strings.rstrip(', ')
        return 'TargetCollection({})'.format(target_strings)



'''
Contains functions that perform depdency parsing.
'''
from collections import defaultdict
from typing import List, Tuple

import networkx as nx
from tweebo_parser import API

from bella.dependency_tokens import DependencyToken
from bella import stanford_tools


class TweeboParser(object):
    '''
    Singleton Class instance
    '''

    instance = None

    def __new__(cls):
        if TweeboParser.instance is None:
            TweeboParser.instance = API(log_errors=True)
        return TweeboParser.instance

    def __getattr__(self, name):
        return getattr(self.instance, name)

    def __setattr__(self, name, value):
        return setattr(self.instance, name, value)


def _to_dependencies_tokens(token_dep_sentence: List[Tuple[str, int]]
                            ) -> List[DependencyToken]:
    '''
    NOTE:
    The DependencyToken allows easy access to all the dependency links for that
    token.

    :param token_dep_sentence: list of tuples that contain (token, linked \
    index token)
    :type token_dep_sentence: list
    :returns: A list of DependencyToken instances one for each tuple/token.
    :rtype: list
    '''

    def dep_search(dep_index, sentence, dep_info):
        '''
        This is a tail recusive function that returns a dictionary denoting
        which index in the sentence relates to the other tokens in the sentence
        and at what dependency depth.

        :param dep_index: The index of the token whos dependecies are being \
        collected
        :param sentence: List of tuples which contain (token, index of \
        head word)
        :param dep_info: default dict whose keys are indexs from the sentence \
        and value is a default dict whose keys are dependency depths and \
        value is the sentence index related to that dependency depth.
        :type dep_index: int
        :type sentence: list
        :type dep_info: defaultdict
        :returns: Default dictionary whose keys are sentence indexs, values \
        are default dictionaries whose keys are dependency depths and value \
        is the associated sentence index to that depth.
        :rtype: defaultdict

        :Example:
        >>> sentence = [('To', -1), ('appear', 0), ('(', -2),
                        ('EMNLP', 1), ('2014', 3)]
        >>> dep_info = defaultdict(lambda: dict())
        >>> print(dep_search(4, sentence, dep_info))
        >>> {0 : {1 : 1, 2 : 3, 3 : 4},
             1 : {1 : 3, 2 : 3},
             3 : {1 : 4},
             4 : {}}
        '''

        head_index = sentence[dep_index][1]
        if head_index == -1 or head_index == -2:
            return dep_info
        prev_dep_info = dep_info[dep_index]
        head_deps = {dep_level + 1: deps
                     for dep_level, deps in prev_dep_info.items()}
        head_deps[1] = dep_index
        dep_info[head_index] = head_deps
        return dep_search(head_index, sentence, dep_info)

    dep_results = defaultdict(lambda: defaultdict(set))
    for index, _ in enumerate(token_dep_sentence):
        dep_result = dep_search(index, token_dep_sentence,
                                defaultdict(lambda: dict()))
        for token_index, dependencies in dep_result.items():
            for dep_level, dependent in dependencies.items():
                dep_results[token_index][dep_level].add(dependent)
    G = nx.Graph()
    for index, token_related_index in enumerate(token_dep_sentence):
        G.add_node(index)
        token, related_index = token_related_index
        related_index = related_index
        if related_index not in {-2, -1}:
            G.add_edge(index, related_index)

    # Convert each of the tokens in the sentence into a dependent token
    # using the results from searching through the dependencies
    dep_tokens = []
    for token_index, token_dep in enumerate(token_dep_sentence):
        token, _ = token_dep
        depth_related = dep_results[token_index]
        token_relations = defaultdict(list)

        connected_indexs = set()
        for _, node_relations in nx.bfs_successors(G, token_index):
            for node_relation in node_relations:
                connected_indexs.add(node_relation)
        if token_index not in connected_indexs:
            connected_indexs.add(token_index)
        connected_indexs = sorted(list(connected_indexs))
        connected_words = []
        for connected_index in connected_indexs:
            connected_word = token_dep_sentence[connected_index][0].strip()
            if connected_index == token_index:
                connected_words.append((connected_word, 'CURRENT'))
            else:
                connected_words.append((connected_word, 'CONNECTED'))
        # Get the tokens relations
        for depth, related_tokens_index in depth_related.items():
            for related_token_index in related_tokens_index:
                related_token = token_dep_sentence[related_token_index][0]
                token_relations[depth].append(related_token)
        dep_tokens.append(DependencyToken(token, token_relations,
                                          connected_words))
    return dep_tokens


def tweebo(texts: List[str]) -> List[DependencyToken]:
    '''
    Given a list of Strings will tokenise, pos tag and then dependecy parse
    the text using `Tweebo <https://github.com/ikekonglp/TweeboParser>`_
    a Tweet specific parser.

    The Tweebo parser cannot handle no strings therefore a special empty string
    symbol is required.

    If one of the texts is an empty String then an empty list will be returned
    for that index of the returned list.

    :param texts: The texts that are to be parsed
    :type text: list
    :returns: A list of of a list of DependencyToken instances. A list per \
    text in the texts argument.
    :rtype: list
    '''

    tweebo_api = TweeboParser()
    texts = [text.replace('\n', ' ') for text in texts]
    processed_texts = tweebo_api.parse_conll(texts)

    dep_texts = []
    for processed_text in processed_texts:
        token_dep_indexs = _convert_conll(processed_text)
        dep_texts.append(_to_dependencies_tokens(token_dep_indexs))
    return dep_texts


def _convert_conll(conll_data: str) -> List[Tuple[str, int]]:
    token_dep_indexs = []
    for line in conll_data.split('\n'):
        if not line:
            continue
        column_data = line.split('\t')
        word = column_data[1].strip()
        # All the word indexs start at 1 not 0.
        # Need to take into account the previous sentence words
        dep_token_index = int(column_data[6]) - 1
        token_dep_indexs.append((word, dep_token_index))
    return token_dep_indexs


def stanford(texts: str) -> List[DependencyToken]:

    dep_texts = []
    for text in texts:
        dep_dicts, tokens_dicts = stanford_tools.dependency_parse(text)
        token_dep_indexs = []
        prev_sent_length = 0
        for sentence, _ in enumerate(tokens_dicts):
            tokens_dict = tokens_dicts[sentence]
            dep_dict = dep_dicts[sentence]
            for i in range(1, len(tokens_dict) + 1):
                word = tokens_dict[i]['word']
                # All the word indexs start at 1 not 0.
                # Need to take into account the previous sentence words
                dep_word_index = (dep_dict[i][1] - 1) + prev_sent_length
                # If True then it is the root word
                if dep_word_index + 1 == prev_sent_length:
                    dep_word_index = -1
                token_dep_indexs.append((word, dep_word_index))
            prev_sent_length += len(tokens_dict)
        dep_texts.append(_to_dependencies_tokens(token_dep_indexs))
    return dep_texts



'''
Functions that allow evaluating model predictions easier, as well as custom
evaluation metrics specific for Target Based Sentiment Analysis e.g. the
:py:func:`bella.evaluation.distinct_sentiment_metrics` can return the
accuracy of the models predictions on samples that have a certain number of
distinct sentiments.

Functions:

1. score -- Given a metric, test data and predictions will evaluate the
   predictions based on the given metric.
2. evaluate_model -- Given a metric, dictionary of dataset names with test data
   and predictions will evaluate the predictions for each dataset.
3. evaluate_models -- Given a metric, dictionary of dataset names with test
   data and a dictionary of models and their predictions will evaluate the
   predictions for each dataset and for each model.
4. distinct_sentiment_metrics -- Custom metric function. It performs the
   `metric_func` on a subset of the test data where the samples have to come
   from sentences that contain a certain number of distinct_sentiments.
'''
from collections import defaultdict
import copy
from typing import Callable, Union, Dict

import numpy as np
import pandas as pd

import bella


def score(metric: Union[Callable[[np.ndarray, np.ndarray], float],
                        Callable[['bella.data_types.TargetCollection',
                                  np.ndarray], float]],
          test_data: 'bella.data_types.TargetCollection',
          predictions: np.ndarray,
          custom_metric: bool = False, **metric_kwargs) -> float:
    '''
    Given a metric, test data and predictions will evaluate the predictions
    based on the given metric.

    :param metric: Metric used to evaluate the predictions e.g.
                   :py:func:`sklearn.metrics.f1_score` or a custom function
                   like :py:func:`bella.evaluation.distinct_sentiment_metrics`
    :param test_data: The test data stored as a
                      :py:class:`bella.data_types.TargetCollection` object.
    :param predictions: Array of predictions for the test data
    :param custom_metric: If the metric being used has come from the Bella
                          package then it is custom and this should be True
                          else use the Default False.
    :param metric_kwargs: Keyword arguments to the metric function.
    :return: The value given back by the metric function.
    '''
    if custom_metric:
        return metric(test_data, predictions, **metric_kwargs)
    return metric(test_data.sentiment_data(), predictions, **metric_kwargs)


def evaluate_model(metric: Union[Callable[[np.ndarray, np.ndarray], float],
                                 Callable[['bella.data_types.TargetCollection',
                                           np.ndarray], float]],
                   dataset_name_data: Dict[str, 'bella.data_types.TargetCollection'],
                   dataset_predictions: Dict[str, np.ndarray],
                   custom_metric: bool = False,
                   **metric_kwargs) -> Dict[str, float]:
    '''
    Given a metric, dictionary of dataset names with test data and predictions
    will evaluate the predictions for each dataset.

    :param metric: Metric used to evaluate the predictions e.g.
                   :py:func:`sklearn.metrics.f1_score` or a custom function
                   like :py:func:`bella.evaluation.distinct_sentiment_metrics`
    :param dataset_name_data: Dictionary of dataset names as keys and the
                              test data as values.
    :param dataset_predictions: Dictionary of dataset names as keys and the
                                predictions as values.
    :param custom_metric: If the metric being used has come from the Bella
                          package then it is custom and this should be True
                          else use the Default False.
    :param metric_kwargs: Keyword arguments to the metric function.
    :return: Dictionary of dataset names as keys and the metric score as
             values.
    '''
    dataset_score = {}
    for dataset_name, test_data in dataset_name_data.items():
        predictions = dataset_predictions[dataset_name]
        dataset_score[dataset_name] = score(metric, test_data, predictions,
                                            custom_metric, **metric_kwargs)
    return dataset_score


def evaluate_models(metric: Union[Callable[[np.ndarray, np.ndarray], float],
                                  Callable[['bella.data_types.TargetCollection',
                                            np.ndarray], float]],
                    dataset_name_data: Dict[str, 'bella.data_types.TargetCollection'],
                    model_dataset_predictions: Dict[str,
                                                    Dict[str, np.ndarray]],
                    custom_metric: bool = False, dataframe: bool = False,
                    **metric_kwargs) -> Union[Dict[str, Dict[str, float]],
                                              'pd.DataFrame']:
    '''
    Given a metric, dictionary of dataset names with test data and
    a dictionary of models and their predictions will evaluate the predictions
    for each dataset and for each model.

    :param metric: Metric used to evaluate the predictions e.g.
                   :py:func:`sklearn.metrics.f1_score` or a custom function
                   like :py:func:`bella.evaluation.distinct_sentiment_metrics`
    :param dataset_name_data: Dictionary of dataset names as keys and the
                              test data as values.
    :param model_dataset_predictions: Dictionary of dictionaries, outer
                                      dictionary is model name and inner
                                      dictionary is dataset names as keys and
                                      predictions as values.
    :param custom_metric: If the metric being used has come from the Bella
                          package then it is custom and this should be True
                          else use the Default False.
    :param dataframe: Whether to return the results as a
                      :py:class:`pandas.DataFrame`. Default False.
    :param metric_kwargs: Keyword arguments to the metric function.
    :return: Dictionary of dictionaries, outer dictionary is the model names,
             the inner is dataset names as keys and predictions as values. If
             `dataframe` argument is True then returns it as a
             :py:class:`pandas.DataFrame` object.
    '''
    model_dataset_score = defaultdict(lambda: dict())
    for model_name, dataset_predictions in model_dataset_predictions.items():
        for dataset_name, predictions in dataset_predictions.items():
            test_data = dataset_name_data[dataset_name]
            result = score(metric, test_data, predictions,
                           custom_metric, **metric_kwargs)
            model_dataset_score[model_name][dataset_name] = result
    if dataframe:
        model_dataset_score = pd.DataFrame(model_dataset_score)
        mean_values = model_dataset_score.mean()
        mean_df = mean_values.to_frame('Mean').T
        return pd.concat((model_dataset_score, mean_df))
    return model_dataset_score


def distinct_sentiment_metrics(test_data: 'bella.data_types.TargetCollection',
                               predictions: np.ndarray,
                               distinct_sentiments: int,
                               metric_func: Callable[[np.ndarray, np.ndarray],
                                                     float],
                               **metric_kwargs) -> float:
    '''
    Custom metric function. It performs the `metric_func` on a subset of the
    test data where the samples have to come from sentences that contain
    a certain number of distinct_sentiments.

    If a sentence has two samples where each target sample is positive this
    would mean that sentence only has one distinct_sentiment. If those two
    samples contained different sentiments then that sentence has two
    distinct_sentiments. This function evaluates predictions on samples that
    have come from sentences with a defined number of distinct_sentiments
    where the evaluation function is the `metric_func` argument over the
    subset of samples based on distinct_sentiments.

    :param test_data: The test data stored as a
                      :py:class:`bella.data_types.TargetCollection` object.
    :param predictions: Array of predictions for the test data
    :param distinct_sentiments: The number of unique sentiments in a sentence.
    :param metric_func: Metric used to evaluate the predictions e.g.
                        :py:func:`sklearn.metrics.f1_score`
    :param metric_kwargs: Keyword arguments to the metric function.
    '''
    test_data_copy = copy.deepcopy(test_data)
    test_data_copy.add_pred_sentiment(predictions)
    test_subset = test_data_copy.subset_by_sentiment(distinct_sentiments)
    true_values = test_subset.sentiment_data()
    subset_predictions = test_subset.sentiment_data(sentiment_field='predicted')
    return metric_func(true_values, subset_predictions, **metric_kwargs)



import numpy as np
from sklearn.base import TransformerMixin
from sklearn.base import BaseEstimator

from bella.neural_pooling import matrix_max

class NeuralPooling(BaseEstimator, TransformerMixin):

    def __init__(self, pool_func=matrix_max):
        self.pool_func = pool_func

    def fit(self, context_word_matrixs, y=None):
        '''Kept for consistnecy with the TransformerMixin'''

        return self

    def fit_transform(self, context_word_matrixs, y=None):
        '''see self.transform'''

        return self.transform(context_word_matrixs)

    def transform(self, context_word_matrixs):
        context_pool_vectors = []
        for context_word_matrix in context_word_matrixs:
            all_contexts = []
            for word_matrix in context_word_matrix:
                all_contexts.append(self.pool_func(word_matrix))
            num_contexts = len(all_contexts)
            vec_size = all_contexts[0].shape[1]
            all_contexts = np.asarray(all_contexts).reshape(num_contexts, vec_size)
            context_pool_vectors.append(all_contexts)
        return context_pool_vectors



from bella.scikit_features import context, debug, join_context_vectors
from bella.scikit_features import lexicon_filter, neural_pooling
from bella.scikit_features import syntactic_context, tokeniser, word_vector



'''
Module contains all of the classes that represent Machine Learning models
that are within `Vo and Zhang 2015 paper \
<https://ijcai.org/Proceedings/15/Papers/194.pdf>`_:

1. :py:class:`bella.models.target.TargetInd` -- Target Indepdent model
2. :py:class:`bella.models.target.TargetDepMinus` -- Target Dependent Minus
   model
3. :py:class:`bella.models.target.TargetDep` -- Target Dependent model
4. :py:class:`bella.models.target.TargetDepPlus` -- Target Dependent Plus model
'''

from typing import Any, List, Callable, Dict

import sklearn
from sklearn.pipeline import FeatureUnion, Pipeline
from sklearn.preprocessing import MinMaxScaler
from sklearn.svm import LinearSVC

import bella
from bella.tokenisers import ark_twokenize
from bella.neural_pooling import matrix_max, matrix_min, matrix_avg
from bella.neural_pooling import matrix_median, matrix_prod, matrix_std
from bella.scikit_features.context import Context
from bella.scikit_features.lexicon_filter import LexiconFilter
from bella.scikit_features.join_context_vectors import JoinContextVectors
from bella.scikit_features.neural_pooling import NeuralPooling
from bella.scikit_features.tokeniser import ContextTokeniser
from bella.scikit_features.word_vector import ContextWordVectors
from bella.models.base import SKLearnModel


class TargetInd(SKLearnModel):
    '''
    Attributes:

    1. model -- Machine learning model. Expects it to be a
       :py:class:`sklearn.pipeline.Pipeline` instance.
    2. fitted -- If the machine learning model has been fitted (default False)
    3. model_parameters -- The parameters that are set in the machine
       learning model. E.g. Parameter could be the tokeniser used.

    Methods:

    1. fit -- Fit the model according to the given training data.
    2. predict -- Predict class labels for samples in X.
    3. probabilities -- The probability of each class label for all samples
       in X.
    4. __repr__ -- Name of the machine learning model.

    Class Methods:

    1. get_parameters -- Transform the given parameters into a dictonary
       that is accepted as model parameters.
    2. get_cv_parameters -- Transform the given parameters into a list of
       dictonaries that is accepted as `param_grid` parameter in
       :py:class:`sklearn.model_selection.GridSearchCV`
    3. name -- -- Returns the name of the model.

    Functions:

    1. save -- Given a instance of this class will save it to a file.
    2. load -- Loads an instance of this class from a file.
    3. pipeline -- Machine Learning model that is used as the base template
       for the model attribute.
    '''

    @classmethod
    def name(cls) -> str:
        return 'Target Independent'

    def __repr__(self) -> str:
        '''
        Name of the machine learning model.
        '''
        return self.name()

    def __init__(self, word_vectors: List['bella.word_vectors.WordVectors'],
                 tokeniser: Callable[[str], List[str]] = ark_twokenize,
                 lower: bool = True, C: float = 0.01,
                 random_state: int = 42,
                 scale: Any = MinMaxScaler()) -> None:
        '''
        :param word_vectors: A list of one or more word vectors to be used as
                             feature vector lookups. If more than one is used
                             the word vectors are concatenated together to
                             create a the feature vector for each word.
        :param tokeniser: Tokeniser to be used e.g. :py:meth:`str.split`
        :param lower: Whether to lower case the words
        :param C: The C value for the :py:class:`sklearn.svm.SVC` estimator
                  that is used in the pipeline.
        :param random_state: The random_state value for the
                             :py:class:`sklearn.svm.SVC` estimator that is
                             used in the pipeline.
        :param scale: How to scale the data before input into the estimator.
                      If no scaling is to be used set this to None.
        '''
        # The parameters here go into the self.get_parameters method
        super().__init__(word_vectors, tokeniser, lower, C, random_state,
                         scale)

    @staticmethod
    def pipeline() -> 'sklearn.pipeline.Pipeline':
        '''
        Machine Learning model that is used as the base template for the model
        attribute.

        :returns: The template machine learning model
        '''
        return Pipeline([
            ('contexts', Context('full')),
            ('tokens', ContextTokeniser()),
            ('word_vectors', ContextWordVectors()),
            ('pool_funcs', FeatureUnion([
                ('max_pipe', Pipeline([
                    ('max', NeuralPooling(matrix_max)),
                    ('join', JoinContextVectors(matrix_median))
                ])),
                ('min_pipe', Pipeline([
                    ('min', NeuralPooling(matrix_min)),
                    ('join', JoinContextVectors(matrix_median))
                ])),
                ('avg_pipe', Pipeline([
                    ('avg', NeuralPooling(matrix_avg)),
                    ('join', JoinContextVectors(matrix_median))
                ])),
                ('prod_pipe', Pipeline([
                    ('min', NeuralPooling(matrix_prod)),
                    ('join', JoinContextVectors(matrix_median))
                ])),
                ('std_pipe', Pipeline([
                    ('min', NeuralPooling(matrix_std)),
                    ('join', JoinContextVectors(matrix_median))
                ]))
            ])),
            ('scale', MinMaxScaler()),
            ('svm', LinearSVC())
        ])

    @classmethod
    def get_parameters(cls,
                       word_vectors: List['bella.word_vectors.WordVectors'],
                       tokeniser: Callable[[str], List[str]] = ark_twokenize,
                       lower: bool = True, C: float = 0.01,
                       random_state: int = 42,
                       scale: Any = MinMaxScaler()) -> Dict[str, Any]:
        '''
        Transform the given parameters into a dictonary that is accepted as
        model parameters

        :param word_vectors: A list of one or more word vectors to be used as
                             feature vector lookups. If more than one is used
                             the word vectors are concatenated together to
                             create a the feature vector for each word.
        :param tokeniser: Tokeniser to be used e.g. :py:meth:`str.split`
        :param lower: Whether to lower case the words
        :param C: The C value for the :py:class:`sklearn.svm.SVC` estimator
                  that is used in the pipeline.
        :param random_state: The random_state value for the
                             :py:class:`sklearn.svm.SVC` estimator that is
                             used in the pipeline.
        :param scale: How to scale the data before input into the estimator.
                      If no scaling is to be used set this to None.
        :return: Model parameters
        '''
        params_dict = {}
        # Add word vectors to Pipeline model
        params_dict = cls._add_to_params_dict(params_dict,
                                              cls._get_word_vector_names(),
                                              word_vectors)
        # Add tokenisers to Pipeline model
        tokenisers_names = [param_name + '__tokeniser'
                            for param_name in cls._get_tokeniser_names()]
        params_dict = cls._add_to_params_dict(params_dict,
                                              tokenisers_names,
                                              tokeniser)
        # Add if the words should be lower cased
        lower_names = [param_name + '__lower'
                       for param_name in cls._get_tokeniser_names()]
        params_dict = cls._add_to_params_dict(params_dict, lower_names, lower)
        # Add how the data should be scaled before going into the SVM
        # If None then it means no scaling happens
        params_dict = cls._add_to_params_dict(params_dict, ['scale'],
                                              scale)
        # Add the C value for the SVM
        params_dict = cls._add_to_params_dict(params_dict, ['svm__C'], C)
        # Add the random state for the SVM
        params_dict = cls._add_to_params_dict(params_dict,
                                              ['svm__random_state'],
                                              random_state)
        return params_dict

    @classmethod
    def get_cv_parameters(cls, word_vectors, tokeniser=[ark_twokenize],
                          lower=[True], C=[0.01], random_state=[42],
                          scale=[MinMaxScaler()]):
        '''
        Transform the given parameters into a list of dictonaries that is
        accepted as `param_grid` parameter in
        :py:class:`sklearn.model_selection.GridSearchCV`

        :param word_vectors: A list of a list of word vectors e.g. [[SSWE()],
                             [SSWE(), GloveCommonCrawl()]].
        :param tokenisers: A list of tokeniser to be used e.g.
                           :py:meth:`str.split`. Default [ark_twokenize]
        :param lowers: A list of bool values which indicate whether to lower
                       case the input words. Default [True]
        :param C: A list of C values for the :py:class:`sklearn.svm.SVC`
                  estimator that is used in the pipeline. Default [0.01]
        :param random_state: A list of random_state values for the
                             :py:class:`sklearn.svm.SVC` estimator that is
                             used in the pipeline. Default [42]
        :param scale: List of scale values. The list can include
                      :py:class:`sklearn.preprocessing.MinMaxScaler` type of
                      clases or None if no scaling is to be used. Default
                      [:py:class:`sklearn.preprocessing.MinMaxScaler`]
        :return: Parameters to explore through cross validation
        '''
        params_list = []
        # Word Vectors
        params_list = cls._add_to_params(params_list, word_vectors,
                                         cls._get_word_vector_names())
        # Tokeniser
        tokenisers_names = [param_name + '__tokeniser'
                            for param_name in cls._get_tokeniser_names()]
        params_list = cls._add_to_params(params_list, tokeniser,
                                         tokenisers_names)
        # Lower
        lower_names = [param_name + '__lower'
                       for param_name in cls._get_tokeniser_names()]
        params_list = cls._add_to_params(params_list, lower, lower_names)
        # C
        params_list = cls._add_to_all_params(params_list, 'svm__C', C)
        # Random State
        params_list = cls._add_to_all_params(params_list, 'svm__random_state',
                                             random_state)
        # Scale
        params_list = cls._add_to_all_params(params_list, 'scale', scale)
        return params_list

    @staticmethod
    def _get_word_vector_names() -> List[str]:
        '''
        Method to be overidden by subclasses as each pipeline will be different
        and will have different parameter names for where the word vectors are
        set.

        :returns: A list of of parameter names where the word vectors are set
                  in the pipeline.
        '''

        return ['word_vectors__vectors']

    @staticmethod
    def _get_tokeniser_names() -> List[str]:
        '''
        Method to be overidden by subclasses as each pipeline will be different
        and will have different parameter names for where tokenisers are
        used.

        :returns: A list of of parameter names where the tokenisers are used
                  in the pipeline.
        '''

        return ['tokens']


class TargetDepMinus(TargetInd):

    @classmethod
    def name(cls) -> str:
        return 'Target Dependent Minus'

    def __repr__(self) -> str:
        '''
        Name of the machine learning model.
        '''
        return self.name()

    def __init__(self, word_vectors: List['bella.word_vectors.WordVectors'],
                 tokeniser: Callable[[str], List[str]] = ark_twokenize,
                 lower: bool = True, C: float = 0.025,
                 random_state: int = 42,
                 scale: Any = MinMaxScaler()) -> None:
        '''
        :param word_vectors: A list of one or more word vectors to be used as
                             feature vector lookups. If more than one is used
                             the word vectors are concatenated together to
                             create a the feature vector for each word.
        :param tokeniser: Tokeniser to be used e.g. :py:meth:`str.split`
        :param lower: Wether to lower case the words
        :param C: The C value for the :py:class:`sklearn.svm.SVC` estimator
                  that is used in the pipeline.
        :param random_state: The random_state value for the
                             :py:class:`sklearn.svm.SVC` estimator that is used
                             in the pipeline.
        :param scale: How to scale the data before input into the estimator.
                      If no scaling is to be used set this to None.
        '''
        # Inherit from SKLearnModel __init__ method
        # The parameters here go into the self.get_parameters method
        super(TargetInd, self).__init__(word_vectors, tokeniser, lower, C,
                                        random_state, scale)

    @staticmethod
    def pipeline() -> 'sklearn.pipeline.Pipeline':
        '''
        Machine Learning model that is used as the base template for the model
        attribute.

        :returns: The template machine learning model
        '''
        return Pipeline([
            ('union', FeatureUnion([
                ('left', Pipeline([
                    ('contexts', Context('left')),
                    ('tokens', ContextTokeniser()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ])),
                ('right', Pipeline([
                    ('contexts', Context('right')),
                    ('tokens', ContextTokeniser()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ])),
                ('target', Pipeline([
                    ('contexts', Context('target')),
                    ('tokens', ContextTokeniser()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ]))
            ])),
            ('scale', MinMaxScaler()),
            ('svm', LinearSVC())
        ])

    @classmethod
    def get_parameters(cls,
                       word_vectors: List['bella.word_vectors.WordVectors'],
                       tokeniser: Callable[[str], List[str]] = ark_twokenize,
                       lower: bool = True, C: float = 0.025,
                       random_state: int = 42,
                       scale: Any = MinMaxScaler()) -> Dict[str, Any]:
        '''
        Transform the given parameters into a dictonary that is accepted as
        model parameters

        :param word_vectors: A list of one or more word vectors to be used as
                             feature vector lookups. If more than one is used
                             the word vectors are concatenated together to
                             create a the feature vector for each word.
        :param tokeniser: Tokeniser to be used e.g. :py:meth:`str.split`
        :param lower: Wether to lower case the words
        :param C: The C value for the :py:class:`sklearn.svm.SVC` estimator
                  that is used in the pipeline.
        :param random_state: The random_state value for the
                             :py:class:`sklearn.svm.SVC` estimator that is used
                             in the pipeline.
        :param scale: How to scale the data before input into the estimator.
                      If no scaling is to be used set this to None.
        :return: Model parameters
        '''
        return super().get_parameters(word_vectors, tokeniser, lower, C,
                                      random_state, scale)

    @classmethod
    def get_cv_parameters(cls, word_vectors, tokeniser=[ark_twokenize],
                          lower=[True], C=[0.025], random_state=[42],
                          scale=[MinMaxScaler()]):
        '''
        Transform the given parameters into a list of dictonaries that is
        accepted as `param_grid` parameter in
        :py:class:`sklearn.model_selection.GridSearchCV`

        :param word_vectors: A list of a list of word vectors e.g. [[SSWE()],
                             [SSWE(), GloveCommonCrawl()]].
        :param tokenisers: A list of tokeniser to be used e.g.
                           :py:meth:`str.split`. Default [ark_twokenize]
        :param lowers: A list of bool values which indicate whether to lower
                       case the input words. Default [True]
        :param C: A list of C values for the :py:class:`sklearn.svm.SVC`
                  estimator that is used in the pipeline. Default [0.025]
        :param random_state: A list of random_state values for the
                             :py:class:`sklearn.svm.SVC` estimator that is
                             used in the pipeline. Default [42]
        :param scale: List of scale values. The list can include
                      :py:class:`sklearn.preprocessing.MinMaxScaler` type of
                      clases or None if no scaling is to be used. Default
                      [:py:class:`sklearn.preprocessing.MinMaxScaler`]
        :return: Parameters to explore through cross validation
        '''
        return super().get_cv_parameters(word_vectors, tokeniser, lower, C,
                                         random_state, scale)

    @staticmethod
    def _get_word_vector_names() -> List[str]:
        '''
        :returns: A list of of parameter names where the word vectors are set
        in the pipeline.
        '''

        return ['union__left__word_vectors__vectors',
                'union__right__word_vectors__vectors',
                'union__target__word_vectors__vectors']

    @staticmethod
    def _get_tokeniser_names() -> List[str]:
        '''
        :returns: A list of of parameter names where the tokenisers are set
        in the pipeline.
        '''

        return ['union__left__tokens',
                'union__right__tokens',
                'union__target__tokens']


class TargetDep(TargetInd):
    '''
    Target-dep model from `Vo and Zhang 2015 paper
    <https://ijcai.org/Proceedings/15/Papers/194.pdf>`_.
    '''

    @classmethod
    def name(cls) -> str:
        return 'Target Dependent'

    def __repr__(self) -> str:
        '''
        Name of the machine learning model.
        '''
        return self.name()

    def __init__(self, word_vectors: List['bella.word_vectors.WordVectors'],
                 tokeniser: Callable[[str], List[str]] = ark_twokenize,
                 lower: bool = True, C: float = 0.01,
                 random_state: int = 42,
                 scale: Any = MinMaxScaler()) -> None:
        '''
        :param word_vectors: A list of one or more word vectors to be used as
                             feature vector lookups. If more than one is used
                             the word vectors are concatenated together to
                             create a the feature vector for each word.
        :param tokeniser: Tokeniser to be used e.g. :py:meth:`str.split`
        :param lower: Wether to lower case the words
        :param C: The C value for the :py:class:`sklearn.svm.SVC` estimator
                  that is used in the pipeline.
        :param random_state: The random_state value for the
                             :py:class:`sklearn.svm.SVC` estimator that is
                             used in the pipeline.
        :param scale: How to scale the data before input into the estimator.
                      If no scaling is to be used set this to None.
        '''
        # Inherit from SKLearnModel __init__ method
        # The parameters here go into the self.get_parameters method
        super(TargetInd, self).__init__(word_vectors, tokeniser, lower, C,
                                        random_state, scale)

    @staticmethod
    def pipeline() -> 'sklearn.pipeline.Pipeline':
        '''
        Machine Learning model that is used as the base template for the model
        attribute.

        :returns: The template machine learning model
        '''
        return Pipeline([
            ('union', FeatureUnion([
                ('left', Pipeline([
                    ('contexts', Context('left')),
                    ('tokens', ContextTokeniser()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ])),
                ('right', Pipeline([
                    ('contexts', Context('right')),
                    ('tokens', ContextTokeniser()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ])),
                ('target', Pipeline([
                    ('contexts', Context('target')),
                    ('tokens', ContextTokeniser()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ])),
                ('full', Pipeline([
                    ('contexts', Context('full')),
                    ('tokens', ContextTokeniser()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ]))
            ])),
            ('scale', MinMaxScaler()),
            ('svm', LinearSVC())
        ])

    @classmethod
    def get_parameters(cls,
                       word_vectors: List['bella.word_vectors.WordVectors'],
                       tokeniser: Callable[[str], List[str]] = ark_twokenize,
                       lower: bool = True, C: float = 0.01,
                       random_state: int = 42,
                       scale: Any = MinMaxScaler()) -> Dict[str, Any]:
        '''
        Transform the given parameters into a dictonary that is accepted as
        model parameters

        :param word_vectors: A list of one or more word vectors to be used as
                             feature vector lookups. If more than one is used
                             the word vectors are concatenated together to
                             create a the feature vector for each word.
        :param tokeniser: Tokeniser to be used e.g. :py:meth:`str.split`
        :param lower: Wether to lower case the words
        :param C: The C value for the :py:class:`sklearn.svm.SVC` estimator
                  that is used in the pipeline.
        :param random_state: The random_state value for the
                             :py:class:`sklearn.svm.SVC` estimator that is
                             used in the pipeline.
        :param scale: How to scale the data before input into the estimator.
                      If no scaling is to be used set this to None.
        :return: Model parameters
        '''
        return super().get_parameters(word_vectors, tokeniser, lower, C,
                                      random_state, scale)

    @classmethod
    def get_cv_parameters(cls, word_vectors, tokeniser=[ark_twokenize],
                          lower=[True], C=[0.01], random_state=[42],
                          scale=[MinMaxScaler()]):
        '''
        Transform the given parameters into a list of dictonaries that is
        accepted as `param_grid` parameter in
        :py:class:`sklearn.model_selection.GridSearchCV`

        :param word_vectors: A list of a list of word vectors e.g. [[SSWE()],
                             [SSWE(), GloveCommonCrawl()]].
        :param tokenisers: A list of tokeniser to be used e.g.
                           :py:meth:`str.split`. Default [ark_twokenize]
        :param lowers: A list of bool values which indicate whether to lower
                       case the input words. Default [True]
        :param C: A list of C values for the :py:class:`sklearn.svm.SVC`
                  estimator that is used in the pipeline. Default [0.01]
        :param random_state: A list of random_state values for the
                             :py:class:`sklearn.svm.SVC` estimator that is
                             used in the pipeline. Default [42]
        :param scale: List of scale values. The list can include
                      :py:class:`sklearn.preprocessing.MinMaxScaler` type of
                      clases or None if no scaling is to be used. Default
                      [:py:class:`sklearn.preprocessing.MinMaxScaler`]
        :return: Parameters to explore through cross validation
        '''
        return super().get_cv_parameters(word_vectors, tokeniser, lower, C,
                                         random_state, scale)

    @staticmethod
    def _get_word_vector_names() -> List[str]:
        '''
        :returns: A list of of parameter names where the word vectors are set
        in the pipeline.
        '''

        return ['union__left__word_vectors__vectors',
                'union__right__word_vectors__vectors',
                'union__target__word_vectors__vectors',
                'union__full__word_vectors__vectors']

    @staticmethod
    def _get_tokeniser_names() -> List[str]:
        '''
        :returns: A list of of parameter names where the tokenisers are set
        in the pipeline.
        '''

        return ['union__left__tokens',
                'union__right__tokens',
                'union__target__tokens',
                'union__full__tokens']


class TargetDepPlus(TargetInd):

    @classmethod
    def name(cls) -> str:
        return 'Target Dependent Plus'

    def __repr__(self) -> str:
        '''
        Name of the machine learning model.
        '''
        return self.name()

    def __init__(self, word_vectors: List['bella.word_vectors.WordVectors'],
                 senti_lexicon: 'bella.lexicons.Lexicon',
                 tokeniser: Callable[[str], List[str]] = ark_twokenize,
                 lower: bool = True, C: float = 0.01,
                 random_state: int = 42,
                 scale: Any = MinMaxScaler()) -> None:
        '''
        :param word_vectors: A list of one or more word vectors to be used as
                             feature vector lookups. If more than one is used
                             the word vectors are concatenated together to
                             create a the feature vector for each word.
        :param senti_lexicon: Sentiment Lexicon to be used for the Left and
                              Right sentiment context (LS and RS).
        :param tokeniser: Tokeniser to be used e.g. :py:meth:`str.split`
        :param lower: Whether to lower case the words
        :param C: The C value for the :py:class:`sklearn.svm.SVC` estimator
                  that is used in the pipeline.
        :param random_state: The random_state value for the
                             :py:class:`sklearn.svm.SVC` estimator that is used
                             in the pipeline.
        :param scale: How to scale the data before input into the estimator.
                      If no scaling is to be used set this to None.
        '''
        # Inherit from SKLearnModel __init__ method
        # The parameters here go into the self.get_parameters method
        super(TargetInd, self).__init__(word_vectors, senti_lexicon, tokeniser,
                                        lower, C, random_state, scale)

    @staticmethod
    def pipeline() -> 'sklearn.pipeline.Pipeline':
        '''
        Machine Learning model that is used as the base template for the model
        attribute.

        :returns: The template machine learning model
        '''
        return Pipeline([
            ('union', FeatureUnion([
                ('left', Pipeline([
                    ('contexts', Context('left')),
                    ('tokens', ContextTokeniser()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ])),
                ('left_s', Pipeline([
                    ('contexts', Context('left')),
                    ('tokens', ContextTokeniser()),
                    ('filter', LexiconFilter()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ])),
                ('right', Pipeline([
                    ('contexts', Context('right')),
                    ('tokens', ContextTokeniser()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ])),
                ('right_s', Pipeline([
                    ('contexts', Context('right')),
                    ('tokens', ContextTokeniser()),
                    ('filter', LexiconFilter()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ])),
                ('target', Pipeline([
                    ('contexts', Context('target')),
                    ('tokens', ContextTokeniser()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ])),
                ('full', Pipeline([
                    ('contexts', Context('full')),
                    ('tokens', ContextTokeniser()),
                    ('word_vectors', ContextWordVectors()),
                    ('pool_funcs', FeatureUnion([
                        ('max_pipe', Pipeline([
                            ('max', NeuralPooling(matrix_max)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('min_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_min)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('avg_pipe', Pipeline([
                            ('avg', NeuralPooling(matrix_avg)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('prod_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_prod)),
                            ('join', JoinContextVectors(matrix_median))
                        ])),
                        ('std_pipe', Pipeline([
                            ('min', NeuralPooling(matrix_std)),
                            ('join', JoinContextVectors(matrix_median))
                        ]))
                    ]))
                ]))
            ])),
            ('scale', MinMaxScaler()),
            ('svm', LinearSVC())
        ])

    @classmethod
    def get_parameters(cls,
                       word_vectors: List['bella.word_vectors.WordVectors'],
                       senti_lexicon: 'bella.lexicons.Lexicon',
                       tokeniser: Callable[[str], List[str]] = ark_twokenize,
                       lower: bool = True, C: float = 0.01,
                       random_state: int = 42,
                       scale: Any = MinMaxScaler()) -> Dict[str, Any]:
        '''
        Transform the given parameters into a dictonary that is accepted as
        model parameters

        :param word_vectors: A list of one or more word vectors to be used as
                             feature vector lookups. If more than one is used
                             the word vectors are concatenated together to
                             create a the feature vector for each word.
        :param senti_lexicon: Sentiment Lexicon to be used for the Left and
                              Right sentiment context (LS and RS).
        :param tokeniser: Tokeniser to be used e.g. :py:meth:`str.split`
        :param lower: Whether to lower case the words
        :param C: The C value for the :py:class:`sklearn.svm.SVC` estimator
                  that is used in the pipeline.
        :param random_state: The random_state value for the
                             :py:class:`sklearn.svm.SVC` estimator that is used
                             in the pipeline.
        :param scale: How to scale the data before input into the estimator.
                      If no scaling is to be used set this to None.
        :return: Model parameters
        '''
        params_dict = super().get_parameters(word_vectors, tokeniser, lower, C,
                                             random_state, scale)
        params_dict = cls._add_to_params_dict(params_dict,
                                              cls._get_word_senti_names(),
                                              senti_lexicon)
        return params_dict

    @classmethod
    def get_cv_parameters(cls,
                          word_vectors: List[List['bella.word_vectors\
                                                        .WordVectors']],
                          senti_lexicon: List['bella.lexicons.Lexicon'],
                          tokeniser=[ark_twokenize],
                          lower=[True], C=[0.01], random_state=[42],
                          scale=[MinMaxScaler()]):
        '''
        Transform the given parameters into a list of dictonaries that is
        accepted as `param_grid` parameter in
        :py:class:`sklearn.model_selection.GridSearchCV`

        :param word_vectors: A list of a list of word vectors e.g. [[SSWE()],
                             [SSWE(), GloveCommonCrawl()]].
        :param senti_lexicon: A list of Sentiment Lexicons to be explored for
                              the Left and Right sentiment context (LS and RS).
                              Default None, use the sentiment lexicons already
                              within the model.
        :param tokenisers: A list of tokeniser to be used e.g.
                           :py:meth:`str.split`. Default [ark_twokenize]
        :param lowers: A list of bool values which indicate whether to lower
                       case the input words. Default [True]
        :param C: A list of C values for the :py:class:`sklearn.svm.SVC`
                  estimator that is used in the pipeline. Default [0.01]
        :param random_state: A list of random_state values for the
                             :py:class:`sklearn.svm.SVC` estimator that is
                             used in the pipeline. Default [42]
        :param scale: List of scale values. The list can include
                      :py:class:`sklearn.preprocessing.MinMaxScaler` type of
                      clases or None if no scaling is to be used. Default
                      [:py:class:`sklearn.preprocessing.MinMaxScaler`]
        :return: Parameters to explore through cross validation
        '''
        params_list = super().get_cv_parameters(word_vectors, tokeniser,
                                                lower, C, random_state, scale)
        params_list = cls._add_to_params(params_list, senti_lexicon,
                                         cls._get_word_senti_names())
        return params_list

    @staticmethod
    def _get_word_vector_names() -> List[str]:
        '''
        :returns: A list of of parameter names where the word vectors are set
        in the pipeline.
        '''

        return ['union__left__word_vectors__vectors',
                'union__left_s__word_vectors__vectors',
                'union__right__word_vectors__vectors',
                'union__right_s__word_vectors__vectors',
                'union__target__word_vectors__vectors',
                'union__full__word_vectors__vectors']

    @staticmethod
    def _get_tokeniser_names() -> List[str]:
        '''
        :returns: A list of of parameter names where the tokenisers are set
        in the pipeline.
        '''

        return ['union__left__tokens',
                'union__left_s__tokens',
                'union__right__tokens',
                'union__right_s__tokens',
                'union__target__tokens',
                'union__full__tokens']

    @staticmethod
    def _get_word_senti_names() -> List[str]:
        '''
        :returns: A list of of parameter names where the sentiment lexicons
        are set in the pipeline.
        '''

        return ['union__left_s__filter__lexicon',
                'union__right_s__filter__lexicon']



'''
Functions that tokenise text and returns the list of tokens all of which are
Strings.

Functions:

1. :py:func:`bella.tokenisers.whitespace` -- tokenises on whitespace.
2. :py:func:`bella.tokenisers.ark_twokenize` -- A Twitter tokeniser from
   `CMU Ark <https://github.com/brendano/ark-tweet-nlp>`_
3. :py:func:`bella.tokenisers.stanford` -- Stanford tokeniser from
   `CoreNLP <https://nlp.stanford.edu/software/tokenizer.html>`_
4. :py:func:`bella.tokenisers.moses` -- Tokeniser used in the
   `moses toolkit <https://github.com/moses-smt>`_
'''
from typing import List

from mosestokenizer import MosesTokenizer
import twokenize

from bella import stanford_tools


def whitespace(text: str) -> List[str]:
    '''
    Tokenises on whitespace.

    :param text: A string to be tokenised.
    :returns: A list of tokens where each token is a String.
    '''

    if isinstance(text, str):
        return text.split()
    raise ValueError(f'The paramter must be of type str not {type(text)}')


def ark_twokenize(text: str) -> List[str]:
    '''
    A Twitter tokeniser from
    `CMU Ark <https://github.com/brendano/ark-tweet-nlp>`_

    This is a wrapper of
    `this <https://github.com/Sentimentron/ark-twokenize-py>`_

    :param text: A string to be tokenised.
    :returns: A list of tokens where each token is a String.
    '''

    if isinstance(text, str):
        return twokenize.tokenizeRawTweetText(text)
    raise ValueError(f'The paramter must be of type str not {type(text)}')


def stanford(text: str) -> List[str]:
    '''
    Stanford tokeniser from
    `CoreNLP <https://nlp.stanford.edu/software/tokenizer.html>`_

    Requires CoreNLP server to be running.

    :param text: A string to be tokenised.
    :returns: A list of tokens where each token is a String.
    '''
    if isinstance(text, str):
        return stanford_tools.tokenise(text)
    raise ValueError(f'The paramter must be of type str not {type(text)}')


class Moses(object):
    '''
    Singleton Class instance
    '''

    instance = None

    def __new__(cls):
        if Moses.instance is None:
            Moses.instance = MosesTokenizer('en')
        return Moses.instance

    def __getattr__(self, name):
        return getattr(self.instance, name)

    def __setattr__(self, name, value):
        return setattr(self.instance, name, value)


def moses(text: str) -> List[str]:
    '''
    Tokeniser used in the `moses toolkit <https://github.com/moses-smt>`_

    This is a wrapper of `this <https://pypi.org/project/mosestokenizer/>`_

    It expects the text not to contain any new lines therefore we split the
    text by new lines and then join the tokens in each line together.

    :param text: A string to be tokenised.
    :returns: A list of tokens where each token is a String.
    '''
    if isinstance(text, str):
        tokeniser = Moses()
        new_line_tokens = [tokeniser(new_line_text)
                           for new_line_text in text.split('\n')]
        return [tokens for new_line in new_line_tokens
                for tokens in new_line]
    raise ValueError(f'The paramter must be of type str not {type(text)}')



'''
Functions that are used throughout the package:

1. :py:func:`bella.helper.read_config` -- Given a file location to a Yaml file
   that stores only keys and relative or full path locations as values. Returns
   those paths as a fully resolved String given the key.
2. :py:func:`download_file` -- Given a file path and the URL address of the
   data, downloads the data from the URL to the file path location.
3. :py:func:`download_model` -- Downloads the specified model trained on the
   dataset from the model zoo and returns it as a trained model to use.
'''
from pathlib import Path

import requests
from ruamel.yaml import YAML

import bella


BELLA_MODEL_DIR = Path.home().joinpath('.Bella', 'Models')


def read_config(key: str, config_file_path: Path) -> str:
    '''
    Given a file location to a Yaml file that stores only keys and relative
    or full path locations as values. Returns those paths as a fully resolved
    String given the key.

    :param key: The key to the Path within the config file.
    :param config_file_path: File path to the YAML configuration file.
    :return: Full path as a String to the relative file location that is
             stored in the key of the config file.
    '''

    config_parent = config_file_path.parent
    with config_file_path.open('r') as config_file:
        yaml = YAML()
        config_data = yaml.load(config_file)
        if key in config_data:
            file_path = config_data[key]
            key_file_path = config_parent.joinpath(file_path)
            return str(key_file_path.resolve())
        else:
            raise ValueError(f'This key {key} does not exist in the '
                             f'config file {config_file_path}')


def download_file(file_path: Path, url: str) -> None:
    '''
    Given a file path and the URL address of the data, downloads the data
    from the URL to the file path location.

    :param file_path: Path to save the downloaded data to.
    :param url: URL location of the data to be downloaded.
    :return: Nothing
    '''
    with file_path.open('wb') as a_file:
        request = requests.get(url, stream=True)
        for chunk in request.iter_content(chunk_size=128):
            a_file.write(chunk)


def download_model(model: 'bella.models.base.BaseModel',
                   dataset_name: str) -> 'bella.models.base.BaseModel':
    '''
    Downloads the specified model trained on the dataset from the
    model zoo and returns it as a trained model to use.

    The model zoo `URL <https://delta.lancs.ac.uk/mooreap/bella-models>`_

    The model zoo came from the following `paper <https://aclanthology.coli.un\
    i-saarland.de/papers/C18-1097/c18-1097>`_ and results for those models
    on the relevant dataset are within the Mass Evaluation section of the paper

    :param model: Class of model you want to download
    :param dataset_name: Name of the dataset that the model has been trained
                         on.
    :return: An instance of the model class you gave as an argument trained
             on the dataset specified.
    '''

    BELLA_MODEL_DIR.mkdir(parents=True, exist_ok=True)
    base_url = 'https://delta.lancs.ac.uk/mooreap/bella-models/raw/master/'

    model_name = model.name()
    model_file_name = f'{model_name} {dataset_name}'
    model_path = BELLA_MODEL_DIR.joinpath(model_file_name)
    if 'LSTM' in model_name:
        meta_data_path = model_path.with_suffix('.pkl')
        model_data_path = model_path.with_suffix('.h5')
        data_paths = [meta_data_path, model_data_path]
        for data_path in data_paths:
            if not data_path.is_file():
                data_url = base_url + data_path.name
                data_url = data_url.replace(' ', '%20')
                download_file(data_path, data_url)

    else:
        if not model_path.is_file():
            model_url = base_url + model_file_name
            model_url = model_url.replace(' ', '%20')
            download_file(model_path, model_url)
    return model.load(model_path)
